# -*- coding: utf-8 -*-

"""
Main file for the Cosmic-CoNN web-based CR detector.
Boning Dong, CY Xu (cxu@ucsb.edu)
"""

import os
import numpy as np
from astropy.io import fits
from skimage.morphology import dilation, square

try:
    from flask_apscheduler import APScheduler
    from flask import (
        Flask,
        request,
        render_template,
        send_from_directory,
        send_file,
        make_response,
        url_for,
        jsonify,
        abort,
        session,
    )
except ImportError:
    raise ImportError(
        "Flask is not installed.\n"
        "Please run `pip install cosmic-conn[webapp]` to install packages for the web app."
    )

from cosmic_conn.web_app.toolkit.astromsg import (
    FloatListPayload,
    ImagePayload,
    PostResponse,
    ThumbnailPayload,
)
from cosmic_conn.web_app.toolkit.file_manager import UploadFileManager, file_recycle_thread
from cosmic_conn.web_app.toolkit.toolkit import find_noise_thumbnail_coords

# from templates.inference import CR_detector
from cosmic_conn.cr_pipeline.utils_img import zscale

# debug
APP_ROOT = os.path.dirname(os.path.abspath(__file__))
ALLOWED_EXTENSIONS = {"fits", "fz"}

app = Flask(__name__)

# app.config["EXPLAIN_TEMPLATE_LOADING"] = True
app.secret_key = "local_instance"

scheduler = APScheduler()
scheduler.init_app(app)
scheduler.start()

# Upload Storage
FILE_RECYCLE_JOB_ID = "file-recycle-id"
FILE_RECYCLE_JOB_INTERVAL = 10

# For local instance, uploads stored at current working directory
# STORAGE_PATH = os.path.join(app.instance_path, "upload_storage")
STORAGE_PATH = os.path.join(os.getcwd(), "instance_temp_storage")
OUTPUT_PATH = os.path.join(os.getcwd(), "cosmic_conn_output")
os.makedirs(OUTPUT_PATH, exist_ok=True)

file_manager = UploadFileManager(STORAGE_PATH)
# scheduler.add_job(
#     id=FILE_RECYCLE_JOB_ID,
#     func=file_recycle_thread,
#     args=[file_manager],
#     trigger='interval',
#     seconds=FILE_RECYCLE_JOB_INTERVAL
# )

# Thumbnail Numbers
THUMBNAIL_PATCH_SIZE = 64
THUMBNAIL_NUMBER = 20


def allowed_file(filename):
    return "." in filename and filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS


@app.route("/")
def index():
    return render_template("upload.html")


@app.route("/process", methods=["POST"])
def process():
    # get user ip address and uuid address
    try:
        user_ip = request.remote_addr
        user_uuid = request.form["uuid"]
        print("user ip: ", user_ip, "user_id: ", user_uuid)
    except:
        abort(400, "Cannot parse user ip or user id.")

    original_file = request.files["file"]

    # https://flask.palletsprojects.com/en/1.1.x/patterns/fileuploads/
    if not original_file or not allowed_file(original_file.filename):
        return abort(400, "Only *.fits or *.fz files are allowed.")

    # create file key based on user's ip  and uuid
    file_key = "{}_{}".format(user_ip, user_uuid).replace(".", "-")
    file_manager.store_uploaded_file(file_key, original_file)
    status, fits_file_path = file_manager.fetch_uploaded_file_path(file_key)
    # print("fetched file path: ", fits_file_path)
    if not status:
        return abort(400, "Save failed or cannot fetch saved image.")

    # model and settings
    opt = app.config['opt']
    cr_model = app.config['cr_model']

    # start detection:
    with fits.open(fits_file_path) as hdul:
        image = None
        try:
            if opt.ext != 'SCI':
                # if user asigned extension, try it first
                image = hdul[opt.ext].data.astype("float32")

            elif hdul[0].data is not None:
                image = hdul[0].data.astype("float32")

            elif hdul[1].data is not None:
                image = hdul[1].data.astype("float32")

            else:
                image = hdul[opt.ext].data.astype("float32")

            # remove Nan is exist
            image = np.nan_to_num(image, copy=False, nan=0.0)

        except:
            raise ValueError(
                f"No valid data found in extention 0, 1 or {opt.ext}, \
                    -e to specify extension name.")

        # CR detection
        cr_probability = cr_model.detect_cr(image, ret_numpy=True)

        # append the CR probability mask and save a copy
        hdul.append(
            fits.CompImageHDU(cr_probability, name="CR_probability")
        )

        # save a FITS copy with CR probability mask
        fits_out_name = "CR_" + original_file.filename
        fits_out_path = os.path.join(OUTPUT_PATH, fits_out_name)
        hdul.writeto(fits_out_path, overwrite=True)

    session["fits_out_path"] = fits_out_path
    session["fits_out_name"] = fits_out_name

    # save the CR array for later
    cr_mask_path = os.path.join(
        STORAGE_PATH, original_file.filename+"_cr.npy")
    np.save(cr_mask_path, cr_probability)
    session["cr_mask_path"] = cr_mask_path

    # flip for correct orientation before drawing
    image = np.flipud(image)
    cr_probability = np.flipud(cr_probability)

    # Construct response to send back
    post_response = PostResponse()
    # Append Image Payload
    post_response.append_payload(ImagePayload(image))
    post_response.append_payload(ImagePayload(cr_probability))

    # Append Thumbnail
    patch = THUMBNAIL_PATCH_SIZE
    binary_threshold = 0.5
    thumbnail_coords = find_noise_thumbnail_coords(
        cr_probability, binary_threshold, patch
    )
    thumbnail_number = min(THUMBNAIL_NUMBER, len(thumbnail_coords))
    thumbnail_payload = ThumbnailPayload(patch)
    for coord in thumbnail_coords[0:thumbnail_number]:
        thumbnail_payload.add_thumbnail_coord(coord)
    post_response.append_payload(thumbnail_payload)

    # Append zscale data
    zscale_list = list(zscale(image))
    post_response.append_payload(FloatListPayload(zscale_list))

    response = make_response(post_response.tobytes(), 200)
    response.headers.set("Content-Type", "application/octet-stream")

    return response


@app.route("/detect_gpu", methods=["POST"])
def detect_gpu():
    # gpu could have different number but cpu is consistent
    gpu_not_found = str(app.config['cr_model'].device) == 'cpu'
    # 1. call gpu function
    # 2. json obj {"gpu_found": true}
    # 3. return response
    response = {'gpu_detected': not gpu_not_found}
    return jsonify(response)


@app.route("/download", methods=["POST"])
def download():
    # when user download, takes two parameters from the front end,
    # process the mask append it after the CR probability mask
    try:
        n_dilation = int(request.form["dilation"])
        threshold = float(request.form["threshold"])
        print("[download] n_dilation: ", n_dilation, "threshold: ", threshold)
    except:
        abort(400, "Cannot parse dilation or threshold.")

    cr_probability = np.load(session["cr_mask_path"], allow_pickle=True)
    # 1. threshold
    cr_binary = cr_probability > threshold
    # 2. dilation
    for i in range(0, n_dilation):
        cr_binary = dilation(cr_binary, square(3))

    with fits.open(session["fits_out_path"]) as hdul:
        cr_binary = cr_binary.astype("uint8")
        hdul.append(fits.CompImageHDU(
            cr_binary, name="CR_mask", uint=True)
        )
        hdul.writeto(session["fits_out_path"], overwrite=True)

    # then show a message "Binary mask appended to a new FITS copy and saved
    #  in "cosmic_conn_output"."
    return send_file(
        session["fits_out_path"], as_attachment=True,
        attachment_filename=session["fits_out_name"]
    )


@app.route("/fitstest", methods=["GET"])
def fitstest():
    print("received fitstest get")

    return make_response(url_for("static", filename="test-fits.fz"), 200)
    # return make_response("", 200)


@app.route("/get_file/<file>")
def get_file(file):
    return send_from_directory(app.config["session_root"], file)


def main(cr_model, opt):
    # # init model based on console arguments
    app.config['opt'] = opt
    app.config['cr_model'] = cr_model

    app.run(host="127.0.0.1", debug=False)


if __name__ == "__main__":
    # from requests import get
    # ip = get("https://api.ipify.org").text

    main()



# -*- coding: utf-8 -*-

"""
Main entry point for the CR labeling pipeline.
`$ bash scripts/reduce_lco_cr.sh` is an example script.
CY Xu (cxu@ucsb.edu)
"""

import os
from functools import partial
import multiprocessing as mp
import astropy.io.fits as fits

try:
    from cosmic_conn.cr_pipeline.lco_cr_reduction import CR_reduction
    from cosmic_conn.cr_pipeline.nres_cr_reduction import NRES_CR_detector
    from cosmic_conn.cr_pipeline.utils_io import align_banzai, is_fits_file
    from cosmic_conn.cr_pipeline.options import argument_parser
except ImportError:
    raise ImportError(
        "Please run `pip install cosmic-conn[develop]` to install all packages for development."
        )


def reduce_LCO_CR_mask(fname, root, opt):
    """
    Takes an hdul of at least 3 consecutive observations from the same night to
    generate a binary CR mask based on a give Signal to Noise Ratio threshold and (optional) a linear CR model

    Outputs a multi-extention fits with 3 SCI frames:
        hdul[1:1+3] are BANZAI reduced SCI frames
        hdul[4] is the valid_mask flags valid CCD area after projection
        hdul[5] is SEP soruce table
        hdul[6:6+3] are CR masks
        hdul[9:9+3] are ignore masks
    """
    fits_path = os.path.join(root, fname)

    try:
        hdul = fits.open(fits_path, memmap=True, lazy_load_hdus=True)
    except:
        with open(opt.log_file, "a") as log:
            msg = f"issue reading {fits_path}\n\n"
            log.write(msg)
            print(msg)

    # use parent directory, avoid duplicates
    opt.root = os.path.split(root)[0]

    exp_time = hdul["SCI"].header["exptime"]

    if exp_time >= opt.min_exptime:
        # init a CR_detector object
        detector = CR_reduction(hdul.copy(), opt)
        ret = detector.generate_CR_masks()

        # if return a string, the detectioned failed and reason is included in the string
        if isinstance(ret, str):
            with open(opt.log_file, "a") as log:
                msg = f"{fname} skipped, reason: {ret}\n\n"
                log.write(msg)
                print(msg)
    else:
        print(f"{fname} skipped, reason: exposure time {exp_time}")

    # free memory
    hdul.close()


def reduce_NRES_CR(fname, root, opt):
    """
    Outputs a combined fits, if the input have 3 frames:
        hdul[0:3] are aligned banzai frames
        hdul[3:6] are corresponding CR masks
    """
    fits_path = os.path.join(root, fname)

    try:
        hdul = fits.open(fits_path, memmap=True, lazy_load_hdus=True)
    except:
        with open(opt.log_file, "a") as log:
            msg = f"issue reading {fits_path}\n"
            log.write(msg)
            print(msg)

    # use parent directory, avoid duplicates
    opt.root = os.path.split(root)[0]

    detector = NRES_CR_detector(hdul.copy(), opt)
    ret = detector.generate_CR_masks()

    # if return a string, the detectioned failed and reason is included in the string
    if isinstance(ret, str):
        with open(opt.log_file, "a") as log:
            msg = f"{fname} skipped, reason: {ret}\n"
            log.write(msg)
            print(msg)

    # free memory
    hdul.close()


if __name__ == "__main__":
    opt = argument_parser()

    # group consecutive ovservations and reproject/align frames
    if not opt.aligned:
        align_banzai(opt)
    """
    after reprojection, read each aligned fits to reduce the CR mask,
    the mask will be appended after the aligned fits and saved as
    a new fits for training, expected final file structure:
        data/banzai_frms
        data/aligned_fits
        data/masked_fits
    """

    assert os.path.exists(opt.data), f"{opt.data} doesn't exist."

    # logging
    log_file = os.path.join(opt.data, "CR_reduction_log.txt")
    opt.log_file = log_file

    # collect already processed to resume work
    processed = []
    for root, dirs, files in os.walk(opt.data):

        # pass fits already detected
        if root.endswith("masked_fits"):
            for f in files:
                if is_fits_file(f):
                    processed.append(f[:30])

        # pass fits already rejected
        if "rejected" in root:
            for f in files:
                processed.append(f[:30])

    # multithread processing
    for root, dirs, files in os.walk(opt.data):
        if root.endswith("aligned_fits"):

            with open(log_file, "a") as log:
                log.write(
                    f"directory: {root}\n"
                    f"SNR thresholds {opt.snr_thres}, {opt.snr_thres_low}\n"
                    f"dilation {opt.dilation}, minimum CR size {opt.min_cr_size}\n"
                    f"{opt.comment}\n\n"
                )

            # avoid temp files generated by Mac's Finder
            files = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]
            files.sort()

            files = [f for f in files if is_fits_file(f)]

            files = [f for f in files if f[:30] not in processed]

            assert len(files) > 0, f"found no valid fits files, check {root}"

            if opt.nres:
                for f in files:
                    reduce_NRES_CR(f, root, opt)
                continue

            # use multiple CPU cores to speed up
            if opt.cpus > 1:
                with mp.Pool(processes=opt.cpus) as pool:
                    detect_partial = partial(
                        reduce_LCO_CR_mask, root=root, opt=opt)
                    pool.map(detect_partial, files)
            else:
                for f in files:
                    reduce_LCO_CR_mask(f, root, opt)



#!/usr/bin/env python

"""Tests for `cosmic_conn` package."""

import os
import shutil
import argparse

from astropy.io import fits


from cosmic_conn.data_utils import download_test_data
from cosmic_conn.cr_pipeline.utils_io import *
from cosmic_conn.reduce_cr import reduce_LCO_CR_mask

DATA_PATH = download_test_data()


def options():
    # init mock parameters for test
    opt = argparse.Namespace()
    opt.data = DATA_PATH
    opt.aligned = False
    opt.nres = False
    opt.snr_thres = 5
    opt.snr_thres_low = 2.5
    opt.dilation = 5
    opt.flood_fill = False
    opt.min_exptime = 99.0
    opt.min_cr_size = 2
    opt.cpus = 1
    opt.no_png = True
    opt.debug = False
    opt.comment = "test_case"

    log_file = os.path.join(opt.data, "CR_reduction_log.txt")
    opt.log_file = log_file

    return opt


class Test_CR_Labeling_Pipeline:
    def test_reproject(self):
        opt = options()
        assert os.path.exists(opt.data), "test_data directory doesn't exist"

        align_banzai(opt)

        for root, dirs, files in os.walk(opt.data):
            if root.endswith("aligned_fits"):
                for f in files:
                    with fits.open(os.path.join(root, f)) as hdu:
                        assert hdu[0].header["frmtotal"] > 0

    def test_cr_reduction(self):
        opt = options()

        for root, dirs, files in os.walk(opt.data):
            if root.endswith("aligned_fits"):
                for f in files:
                    reduce_LCO_CR_mask(f, root, opt)

        for root, dirs, files in os.walk(opt.data):
            if root.endswith("masked_fits"):
                for f in files:
                    with fits.open(os.path.join(root, f)) as hdu:
                        cr_mask = hdu["CR"].data
                        assert np.sum(cr_mask) > 0

        # clean up after test
        # shutil.rmtree(DATA_PATH)



"""
Main file to define the Cosmic-CoNN pytorch class.
CY Xu (cxu@ucsb.edu)
"""

import os
import math
import random

import torch
import torch.nn as nn
import torch.nn.init as init
import torch.optim as optim


# model import
from cosmic_conn.dl_framework.utils_ml import (
    clean_large,
    memory_check,
    modulus_boundary_crop,
    subtract_sky,
    median_weighted_bce,
    tensor2np,
    DiceLoss,
    remove_nan
)
from cosmic_conn.dl_framework.unet import UNet_module
from cosmic_conn.cr_pipeline.utils_img import save_as_png


class Cosmic_CoNN(nn.Module):
    def __init__(self):
        super(Cosmic_CoNN, self).__init__()

    def name(self):
        return "Unsupervised_CR"

    def initialize(self, opt):
        # torch.cuda.set_device('cuda')
        # torch.device('cuda')
        self.opt = opt
        self.best_name = "epoch_0"
        self.last_best_model = None
        self.best_valid_loss = float("inf")
        self.vis = []

        if torch.cuda.is_available():
            self.device = torch.device("cuda")
            print("...GPU found, yeah!")
        else:
            self.device = torch.device("cpu")
            print("...GPU or CUDA not detected, using CPU (slower). ")
            print("...training on CPU is not recommended.")

        self.model_dir = os.path.join(opt.expr_dir, "models")
        self.valid_dir = os.path.join(opt.expr_dir, "validation")

        if not opt.load_model:
            os.makedirs(self.valid_dir, exist_ok=True)
            os.makedirs(self.model_dir, exist_ok=True)

        self.build_models(opt)

    def build_models(self, opt):
        torch.set_default_tensor_type(torch.FloatTensor)
        torch.set_default_dtype(torch.float32)

        # group norm parameters
        norm_setting = [opt.n_group, opt.gn_channel, opt.no_affine]

        # the network is defined here
        self.network = nn.DataParallel(
            UNet_module(
                n_channels=1,
                n_classes=1,
                hidden=opt.hidden,
                norm=opt.norm,
                norm_setting=norm_setting,
                conv_type=opt.conv_type,
                down_type=opt.down_type,
                up_type=opt.up_type,
                deeper=opt.deeper,
            )
        )

        self.network.to(self.device)

        # count trainlearnableable parameteres
        # model_parameters = filter(lambda p: p.requires_grad, self.network.parameters())
        # total_params = sum([np.prod(p.size()) for p in model_parameters])
        # print(f"Total learnable parameters {total_params}")

        if opt.mode == "train":

            # init common loss functions
            if opt.loss == "bce":
                self.criterion = nn.BCELoss()

            elif opt.loss == "median_bce":
                self.criterion = median_weighted_bce
                self.bce = nn.BCELoss()

            elif opt.loss == "mse":
                self.criterion = nn.MSELoss()

            elif opt.loss == "dice":
                self.criterion = DiceLoss

            else:
                raise ValueError('Invalid loss function, choose from \
                                bce, median_bce, mse, dice')

            # optimizer
            params = self.network.parameters()

            self.optimizer = optim.Adam(params, lr=opt.lr)
            # self.optimizer = optim.SGD(self.network.parameters(), lr=opt.lr)

            # init network
            self.apply(self.weights_init("kaiming"))

            # Multi Step Scheduler
            if opt.milestones[0] == '0':
                milestones = []
            else:
                milestones = [int(i) for i in opt.milestones]

            self.lr_scheduler = optim.lr_scheduler.MultiStepLR(
                self.optimizer, milestones=milestones, gamma=0.1
            )

            # only import tensorboard for training
            try:
                from torch.utils.tensorboard import SummaryWriter
            except ImportError:
                raise ImportError(
                    "Please run `pip install cosmic-conn[develop]` to install all packages for development."
                )

        # Inference or continue training

        if opt.mode == "inference":
            # inference only

            if len(opt.load_model) == 0:
                raise ValueError("Please specify the path to load model file -m.")

            checkpoint = torch.load(opt.load_model, map_location=self.device)

            if opt.load_model.endswith("pth"):
                self.network.load_state_dict(checkpoint)
            else:
                self.network.load_state_dict(checkpoint["state_dict_mask"])

            # use available memory to determine detection method
            self.full_image_detection = memory_check(self.device)

        elif opt.mode == "train" and opt.continue_train:
            # initialize with previously saved model, but new optimizer
            if opt.load_model:
                checkpoint_path = opt.load_model

            elif opt.continue_epoch > 0:
                checkpoint_path = os.path.join(
                    self.model_dir, f"cr_checkpoint_{opt.continue_epoch}.pth.tar"
                )

            else:
                raise ValueError(
                    "To continue traininng, must provid checkpoint directory in \
                    opt.continue_train or model path in opt.load_model"
                )

            checkpoint = torch.load(checkpoint_path, map_location=self.device)

            if opt.load_model:
                if opt.load_model.endswith("pth"):
                    self.network.load_state_dict(checkpoint)
                else:
                    self.network.load_state_dict(checkpoint["state_dict_mask"])

            # continue with previously saved model and optimizer
            elif opt.continue_epoch > 0:
                self.network.load_state_dict(checkpoint["state_dict_mask"])
                self.optimizer.load_state_dict(checkpoint["optimizer_mask"])

                # fast forward scheduler to adjust LR rate properly
                for i in range(opt.continue_epoch):
                    self.lr_scheduler.step()
                print(f'LR scheduler fastforward to {opt.continue_epoch}')

            self.writer = SummaryWriter(
                log_dir=str(opt.expr_dir),
                filename_suffix=f"continue_{opt.continue_epoch}",
                purge_step=opt.continue_epoch + 1,
            )

        else:
            # start new training
            self.writer = SummaryWriter(log_dir=str(opt.expr_dir))

    def forward(self, noisy):
        return self.network(noisy)

    def detect_full_image(self, image):
        # try full image detecction first
        mask = torch.zeros_like(image)
        frame_ = modulus_boundary_crop(image, modulo=16)
        tensor = self.dim_shift(frame_).to(self.device)
        tensor -= tensor.median()
        pdt = self.network(tensor).squeeze()
        mask[: pdt.shape[0], : pdt.shape[1]] = pdt
        return mask

    def detect_image_stamps(self, image, verbose=True):
        # if not enough memory, detect in smaller stamps
        mask = None
        stamp_sizes = [1024, 512, 256]

        for stamp in stamp_sizes:
            try:
                if verbose:
                    print(
                        f"Slicing image to {stamp}x{stamp} stamps...",
                    )
                torch.cuda.empty_cache()

                mask = clean_large(
                    image, self.network, patch=stamp, overlap=0
                )
            except:
                print(f"...{stamp}x{stamp} stamp won't fit into memory.")

            if mask is not None:
                break

        assert mask is not None, f"...detection failed.\n Memory too small?"
        return mask

    def detect_cr(self, image, ret_numpy=True):
        # replace NaN with 0.0 if exist
        image = remove_nan(image)

        # numpy array -> tensor
        if not isinstance(image, torch.Tensor):
            image = torch.tensor(image).float()

        # no gradient saved during inference
        with torch.no_grad():
            if self.full_image_detection:
                try:
                    # on CPU, instance might be killed without raising error
                    mask = self.detect_full_image(image)
                except:
                    mask = self.detect_image_stamps(image)
            else:
                mask = self.detect_image_stamps(image)

        if ret_numpy:
            return tensor2np(mask)
        else:
            return mask

    # @profile
    def update_model(self, data, epoch):
        self.epoch = epoch
        self.optimizer.zero_grad()

        # special case, train in eval() for batch norm
        if self.opt.norm == "batch" and epoch > self.opt.eval_epoch:
            self.eval()
        else:
            self.train()

        # dynamically calculate the imbalance alpha for median weighted BCE
        if epoch <= 1:
            self.imbalance_alpha = 0.0
        else:
            self.imbalance_alpha = min(epoch / self.opt.imbalance_alpha, 1.0)

        # forward and backward propogate one sample at a time, no update
        train_loss = self.forward_pass(data)

        train_loss.backward()
        self.optimizer.step()

        return train_loss

    def forward_pass(self, data):

        # mask_src is bad_mask when it's HST data
        noisy, mask_ref, medians, mask_ignore, filenames = self.prep_data(data)

        mask_pdt = None

        # 1) learn reference mask
        mask_pdt = self.network(noisy)

        # original deepCR implementation ignored gradients from ignore mask
        if "hst" in filenames[0]:
            mask_pdt = mask_pdt * (1 - mask_ignore)
            mask_ref = mask_ref * (1 - mask_ignore)

        if self.opt.loss == "median_bce":

            if self.imbalance_alpha < 1.0:
                loss_ref = self.criterion(
                    mask_pdt, mask_ref, medians, self.imbalance_alpha
                )
            else:
                loss_ref = self.bce(mask_pdt, mask_ref)

        else:
            loss_ref = self.criterion(mask_pdt, mask_ref)

        # for debug, reivew training images
        # self.vis.append([
        #     tensor2np(noisy.squeeze()[0]),
        #     tensor2np(medians.squeeze()[0]),
        #     tensor2np(mask_ref.squeeze()[0]),
        #     tensor2np(mask_pdt.squeeze()[0]), filenames[0]
        # ])

        return loss_ref

    def prep_data(self, data, validate=False):
        """
        data input shape: (batch, patch, frm, h, w)
        validate input shape: (1, frm, h, w)
        output shape (batch * patch, 1, h, w)
        """
        if len(data) == 5:
            frames, cr_masks, ignore_masks, medians, filenames = data
        else:
            # ignore_mask is not used for training for LCO data
            frames, cr_masks, ignore_masks, filenames = data

        # send to gpu for faster median frame calculation
        frames = frames.to(self.device)

        if isinstance(filenames[0], list) or isinstance(filenames[0], tuple):
            filenames = [item for sublist in filenames for item in sublist]

        frames = subtract_sky(frames, remove_negative=False)

        # HST data, no need for following steps
        if "hst" in filenames[0]:
            batch, frm, h, w = frames.shape
            patch = 1
            # reshape to 4 dimensions
            frames = frames.view(-1, 1, h, w)
            cr_masks = cr_masks.view(-1, 1, h, w)

            ignore_masks = ignore_masks.view(-1, 1, h, w)
            medians = medians.view(-1, 1, h, w)

        elif "lco" in filenames[0] or "nres" in filenames[0]:
            # extra processing needed fro LCO data
            batch, frm, h, w = frames.shape

            # calculate median on GPU is faster, -3 is the frame axis in both cases
            medians = torch.median(frames, dim=-3)[0].view(-1, 1, h, w)

            if validate:
                frames = frames.view(-1, 1, h, w)
                cr_masks = cr_masks.view(-1, 1, h, w)
                ignore_masks = ignore_masks.view(-1, 1, h, w)
                ignore_masks[ignore_masks > 0.0] = 1.0

                if len(filenames) != frames.shape[0]:
                    filenames = filenames * frames.shape[0]

            else:
                # extract same frame for both image and cr_mask
                frames, cr_masks = self.extract_random_frame(frames, cr_masks)
                ignore_masks = None

        cr_masks = cr_masks.to(self.device)
        medians = medians.to(self.device)

        if ignore_masks is not None:
            ignore_masks = ignore_masks.to(self.device)

        return [frames, cr_masks, medians, ignore_masks, filenames]

    def extract_random_frame(self, frames, cr_masks):
        # randomly pick one frame from a sequence of three frames
        b, c, h, w = frames.shape
        idx = random.randint(0, c - 1)

        frames = [frames[i, idx] for i in range(b)]
        cr_masks = [cr_masks[i, idx] for i in range(b)]

        frames = torch.stack(frames).view(-1, 1, h, w)
        cr_masks = torch.stack(cr_masks).view(-1, 1, h, w)

        return frames, cr_masks

    # @profile
    def validate(self, valid_loader, epoch, source):
        loss_dict = {}
        loss_key = f"loss/valid_cr_loss_{source}"
        self.eval()
        self.epoch_dir = os.path.join(self.valid_dir, f"epoch_{epoch}")

        # DICE evaluation
        with torch.no_grad():
            # load each fits file
            for i, data in enumerate(valid_loader):

                frames, cr_masks, medians, ignore_masks, filenames = self.prep_data(
                    data, validate=True
                )

                b, c, h, w = frames.shape
                pdt_masks = torch.zeros_like(frames)

                for j in range(b):
                    pdt = self.detect_image_stamps(frames[j].squeeze(), False)
                    pdt_masks[j] = pdt.unsqueeze(dim=0)

                # pdt_masks = self.detect_image_stamps(frames)

                # to correct the training set bias for neutral evaluation
                pdt_masks = pdt_masks * (1 - ignore_masks)
                cr_masks = cr_masks * (1 - ignore_masks)

                # validation loss might be different from training
                # so different training loss could be evaluated at the same standard
                cr_loss = DiceLoss(pdt_masks, cr_masks)

                loss_dict[loss_key] = loss_dict.get(loss_key, 0)\
                    + cr_loss.item()

            # save prediction png for preview
            for j in range(b):
                median = medians.squeeze(
                )[j] if source == "hst" else medians.squeeze()

                self.vis.append(
                    [
                        tensor2np(frames.squeeze()[j]),
                        tensor2np(median),
                        tensor2np(cr_masks.squeeze()[j]),
                        tensor2np(pdt_masks.squeeze()[j]),
                        filenames[j],
                    ]
                )

        # normalize to 1.
        loss_dict[loss_key] = loss_dict[loss_key] / len(valid_loader)

        # sum all losses for a mixed model
        loss_dict["loss/valid_cr_loss"] = (
            loss_dict.get("loss/valid_cr_loss", 0) + loss_dict[loss_key]
        )

        self.loss_dict = loss_dict
        return loss_dict

    def save_checkpoint(self, epoch):

        # save best checkpoing for continue training
        state = {
            "epoch": epoch,
            "state_dict_mask": self.network.state_dict(),
            "optimizer_mask": self.optimizer.state_dict(),
        }

        # save best model for inference
        model_path = f"{self.model_dir}/cr_checkpoint_{epoch}.pth.tar"
        torch.save(state, model_path)

        return model_path

    def get_current_visuals(self, epoch):
        os.makedirs(self.epoch_dir, exist_ok=True)
        root = self.epoch_dir

        for vis in self.vis:
            frm, med, msk_ref, msk_pdt, fname = vis
            text = self.opt.comment if "lco" in fname else None

            _, z1z2 = save_as_png(
                med, root, fname, file_type="1_median", zscale=True, grid=True,
            )

            save_as_png(
                frm,
                root,
                fname,
                file_type="2_noisy",
                zscale=True,
                z1z2=z1z2,
                grid=True,
            )

            save_as_png(msk_ref, root, fname,
                        file_type="4_mask_ref", grid=True)

            save_as_png(msk_pdt, root, fname,
                        file_type="5_mask_pdt", grid=True)

        # reset global vis list after all files saved
        self.vis = []

    def update_learning_rate(self, epoch):

        # log LR first, real current epoch LR
        lr_mask = self.optimizer.param_groups[0]["lr"]
        self.writer.add_scalar("para/lr_mask", lr_mask, epoch)

        self.lr_scheduler.step()

    def weights_init(self, init_type="gaussian"):
        def init_fun(m):
            classname = m.__class__.__name__
            if (
                classname.find("Conv") == 0 or classname.find("Linear") == 0
            ) and hasattr(m, "weight"):
                # print m.__class__.__name__
                if init_type == "gaussian":
                    init.normal_(m.weight.data, 0.0, 0.02)
                elif init_type == "xavier":
                    init.xavier_normal_(m.weight.data, gain=math.sqrt(2))
                elif init_type == "kaiming":
                    init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
                elif init_type == "orthogonal":
                    init.orthogonal_(m.weight.data, gain=math.sqrt(2))
                elif init_type == "default":
                    pass
                else:
                    assert 0, "Unsupported initialization: {}".format(
                        init_type)
                if hasattr(m, "bias") and m.bias is not None:
                    init.constant_(m.bias.data, 0.0)

        return init_fun

    # tensorboard logging
    def log_value(self, key, value, epoch):
        self.writer.add_scalar(key, value, epoch)

    def log_dict(self, loss_dict, epoch):
        for key, value in loss_dict.items():
            self.writer.add_scalar(key, value, epoch)

    def dim_shift(self, tensor):
        tensor = tensor.squeeze()

        if len(tensor.shape) == 4:
            return tensor
        elif len(tensor.shape) == 3:
            return tensor.unsqueeze(dim=1)
        elif len(tensor.shape) == 2:
            return tensor.unsqueeze(dim=0).unsqueeze(dim=0)



#!/usr/bin/env python

"""Tests for `cosmic_conn` package."""

import os
import shutil
import numpy as np
from astropy.io import fits

from cosmic_conn.inference_cr import init_model
from cosmic_conn.data_utils import download_test_data, is_fits_file


PREDICT_DIR = "cosmic_conn_output"
TEST_FILE = download_test_data()
DATA_PATH = os.path.join(TEST_FILE, "banzai_frms")


def inference_on_test_data(opt, model):
    for root, dirs, files in os.walk(opt.input):
        if PREDICT_DIR in root:
            continue

        files = [f for f in files if is_fits_file(f)]

        for f in files:
            with fits.open(os.path.join(root, f)) as hdul:
                frame = hdul["SCI"].data.astype("float32")
                pdt_cosmic = model.detect_cr(frame, ret_numpy=True)
                assert np.sum(pdt_cosmic) > 0.0
                assert np.max(pdt_cosmic) <= 1.0
                assert np.min(pdt_cosmic) >= 0.0

                hdul.append(fits.CompImageHDU(
                    pdt_cosmic, name=f"CR_probability"))

                # a new copy is saved in the prediction directory
                out_dir = os.path.join(root, PREDICT_DIR)
                os.makedirs(out_dir, exist_ok=True)

                fname = os.path.join(out_dir, f"CR_{f}")
                hdul.writeto(fname, overwrite=True)

    # clean up after test
    shutil.rmtree(DATA_PATH)


class Test_model_inference:
    def test_lco_imaging(self):
        model = init_model("ground_imaging")
        model.opt.input = DATA_PATH
        print(f"trained models loaded from {model.opt.load_model}")

        inference_on_test_data(model.opt, model)



# -*- coding: utf-8 -*-

"""
Main entry point for the Cosmic-CoNN deep learning model training.
`$ bash scripts/train_lco.sh` is an example script.
CY Xu (cxu@ucsb.edu)
"""

import sys
import random
import time
import numpy as np

import torch
import torch.backends.cudnn as cudnn

try:
    from cosmic_conn.dl_framework.options import TrainingOptions
    from cosmic_conn.dl_framework.dataloader import CreateDataLoader
    from cosmic_conn.dl_framework.cosmic_conn import Cosmic_CoNN
except ImportError:
    raise ImportError(
        "Please run `pip install cosmic-conn[develop]` to install all packages for development."
        )


def setup_seed(seed):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def search_best_seed(opt, loss_key, candidates=30):
    loss_best = float("inf")
    model_best_path = None
    seed_best = 0

    _, _, valid_loader = CreateDataLoader(opt)

    for i in range(candidates):
        randseed = np.random.randint(0, 2 ** 32)
        setup_seed(randseed)

        model = Cosmic_CoNN()
        model.initialize(opt)
        loss_dict = model.validate(valid_loader, epoch=0, source=opt.model)

        if loss_dict[loss_key] <= loss_best:
            loss_best = loss_dict[loss_key]
            model_best_path = model.save_checkpoint(epoch=0)
            seed_best = randseed
            print("best model")

        del model
        print(f"randseed {randseed}, loss {loss_dict[loss_key]}")

    opt.load_model = model_best_path

    return opt, seed_best


def main():
    # get training arguments/options
    train_options = TrainingOptions()
    opt = train_options.parse()
    loss_key = "loss/valid_cr_loss"

    # deterministic flags for model reproducibility
    cudnn.deterministic = True
    cudnn.benchmark = False

    if opt.random_seed:
        # https://github.com/pytorch/pytorch/issues/12873
        # cudnn.enabled = True
        # cudnn.benchmark = True
        opt, randseed = search_best_seed(opt, loss_key, 30)
    else:
        randseed = opt.seed

    # seed everything before instances are created
    setup_seed(randseed)

    data_loader, train_set, valid_data_loader = CreateDataLoader(opt)
    model = Cosmic_CoNN()
    model.initialize(opt)

    # logging
    with open(opt.log_file, "a") as log_file:
        now = time.strftime("%c")
        log_file.write(f"random seed: {randseed}")
        log_file.write(
            f"\n================ Training Loss ({now}) ================\n")

    # epoch loop, continue_epoch is 0 by default
    for e in range(0, opt.epoch + 1, 1):
        tic = time.perf_counter()
        train_loss = 0.0
        loss_dict = {}

        # load NEXT subset of data to memory (parallel)
        train_set.sample_subset()

        # repeat sampling process so data dataloader
        # have same randomness when continue from a previous job
        if e <= opt.continue_epoch:
            continue
        else:
            log_file = open(opt.log_file, "a")
            log_file.write(f"epoch {e}\n")
            print(f"\nepoch {e}")

        # sample loop
        for i, data in enumerate(data_loader):
            loss = model.update_model(data, e)
            train_loss += loss.item()

        train_loss /= len(data_loader)
        toc = time.perf_counter()

        # logging
        model.log_value("loss/train_loss", train_loss, e)
        log = f"train loss {round(train_loss, 10)}, time {round(toc-tic)}s"
        log_file.write(log + "\n")
        print(log)

        # validation
        if e % opt.validate_freq == 0:
            loss_dict = model.validate(valid_data_loader, e, source=opt.model)
            model.log_dict(loss_dict, e)

            # best validation loss and save model
            if loss_dict[loss_key] <= model.best_valid_loss:
                model.best_valid_loss = loss_dict[loss_key]
                model.get_current_visuals(e)
                # self.test_model(epoch)

                log = f"new best model at epoch {e}"
                log_file.write(log + "\n")
                print(log)

            # save checkpoint every epoch
            model.best_model_path = model.save_checkpoint(e)

            # logging
            valid_loss = round(loss_dict[loss_key], 10)
            log = f"valid loss {valid_loss}, time {round(time.perf_counter()-toc)}s"
            log_file.write(log + "\n")
            print(log)

        # update LR at end of each epoch
        model.update_learning_rate(e)
        log_file.write("\n")
        log_file.close()
        print()

    sys.exit("Finish training")


if __name__ == "__main__":
    torch.set_num_threads(1)
    main()



"""
Defies a versatile U-Net model building procedure.
CY Xu (cxu@ucsb.edu)
"""
import torch
import torch.nn as nn


class DoubleConv(nn.Module):
    def __init__(self, in_ch, out_ch, norm, norm_setting):
        super(DoubleConv, self).__init__()

        # group norm fixed/dynamic group numbers
        group, channel, no_affine = norm_setting
        affine = not no_affine

        if group == 0 and channel > 0:
            # dynamically divide groups based on channel number
            n_group = max(1, int(out_ch // channel))
        else:
            # fixed group number
            n_group = group

        norms = nn.ModuleDict(
            {
                "batch": nn.BatchNorm2d(
                    out_ch, momentum=0.005, affine=True, track_running_stats=True
                ),
                "group": nn.GroupNorm(n_group, out_ch, affine=affine),
                "instance": nn.InstanceNorm2d(out_ch, affine=affine),
            }
        )

        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, 1, padding=1),
            norms[norm],
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, 1, padding=1),
            norms[norm],
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class ResBlock(nn.Module):
    def __init__(self, in_ch, out_ch, norm, norm_setting):
        super(ResBlock, self).__init__()

        # group norm fixed/dynamic group numbers
        group, channel, no_affine = norm_setting

        if group == 0 and channel > 0:
            # dynamically divide groups based on channel number
            n_group = max(1, int(out_ch // channel))
        else:
            # fixed group number
            n_group = group

        norms = nn.ModuleDict(
            {
                "batch": nn.BatchNorm2d(out_ch, momentum=0.005),
                "group": nn.GroupNorm(n_group, out_ch),
                "instance": nn.InstanceNorm2d(out_ch, affine=False),
            }
        )

        self.residual = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size=1,
                      stride=1, bias=True), norms[norm]
        )

        self.double_conv = DoubleConv(in_ch, out_ch, norm, norm_setting)
        self.relu = nn.ReLU()

    def forward(self, x):
        identity = self.residual(x)
        out = self.double_conv(x)
        out = self.relu(out + identity)
        return out


class InConv(nn.Module):
    def __init__(self, in_ch, out_ch, norm, norm_setting):
        super(InConv, self).__init__()
        self.conv = DoubleConv(in_ch, out_ch, norm, norm_setting)

    def forward(self, x):
        x = self.conv(x)
        return x


class Down(nn.Module):
    def __init__(self, in_ch, out_ch, norm, norm_setting, conv_type, down_type):
        super(Down, self).__init__()

        downs = nn.ModuleDict(
            {
                "maxpool": nn.MaxPool2d(2),
                "avgpool": nn.AvgPool2d(2),
                "stride": nn.Conv2d(in_ch, in_ch, 3, 2, padding=1),
            }
        )

        convs = nn.ModuleDict(
            {
                "unet": DoubleConv(in_ch, out_ch, norm, norm_setting),
                "resnet": ResBlock(in_ch, out_ch, norm, norm_setting),
            }
        )

        self.down_cov = nn.Sequential(downs[down_type], convs[conv_type])

    def forward(self, x):
        x = self.down_cov(x)
        return x


class Up(nn.Module):
    def __init__(self, in_ch, out_ch, norm, norm_setting, up_type):
        super(Up, self).__init__()

        if up_type == "deconv":
            self.up_conv = nn.ConvTranspose2d(in_ch, out_ch, 2, stride=2)

        elif up_type == "upscale":
            self.up_conv = nn.Sequential(
                nn.Upsample(scale_factor=2, mode="bilinear",
                            align_corners=False),
                nn.Conv2d(in_ch, out_ch, 3, 1, padding=1),
            )

        else:
            raise ValueError(
                f"unknown up_type {up_type}, acceptable transconv, upscale"
            )

        self.conv = DoubleConv(in_ch, out_ch, norm, norm_setting)

    def forward(self, x1, x2):
        x1 = self.up_conv(x1)
        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x


class OutConv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, 1)

    def forward(self, x):
        x = self.conv(x)
        return x


class UNet_module(nn.Module):
    def __init__(
        self,
        n_channels,
        n_classes,
        hidden,
        norm,
        norm_setting,
        conv_type,
        down_type,
        up_type,
        deeper,
    ):
        super(UNet_module, self).__init__()
        self.deeper = deeper

        self.inc = InConv(n_channels, hidden, norm, norm_setting)

        self.down1 = Down(hidden, hidden * 2, norm,
                          norm_setting, conv_type, down_type)

        self.down2 = Down(hidden * 2, hidden * 4, norm,
                          norm_setting, conv_type, down_type)

        if deeper:
            self.down3 = Down(
                hidden * 4, hidden * 8, norm, norm_setting, conv_type, down_type
            )
            self.up3 = Up(hidden * 8, hidden * 4, norm, norm_setting, up_type)

        self.up2 = Up(hidden * 4, hidden * 2, norm, norm_setting, up_type)

        self.up1 = Up(hidden * 2, hidden, norm, norm_setting, up_type)

        self.outc = OutConv(hidden, n_classes)

    def forward(self, x):
        x1 = self.inc(x)

        x2 = self.down1(x1)
        x3 = self.down2(x2)

        if self.deeper:
            x4 = self.down3(x3)
            x = self.up3(x4, x3)
            x = self.up2(x, x2)
            x = self.up1(x, x1)
        else:
            x = self.up2(x3, x2)
            x = self.up1(x, x1)

        x = self.outc(x)
        x = torch.sigmoid(x)
        return x



# -*- coding: utf-8 -*-

"""
Note: When deploying the application to a online webserver, this method may need to be changed.
Boning Dong, CY Xu (cxu@ucsb.edu)
"""

import os
from datetime import datetime
import shutil
from typing import Tuple
import time

FILE_PERSIST_TIME_MINS = 0.1


def file_recycle_thread(file_manager):
    print("recycling uploaded files ...")
    keys_to_delete = []
    for key, record in file_manager.upload_file_records.items():
        dt = record.create_time - datetime.now()
        if dt.seconds >= FILE_PERSIST_TIME_MINS * 60:
            keys_to_delete.append(key)

    for key in keys_to_delete:
        # delete the file
        record = file_manager.upload_file_records[key]
        os.remove(record.file_path)
        # delete the record
        del file_manager.upload_file_records[key]


class UploadFileRecord:
    def __init__(self, file_path):
        self.file_path = file_path
        self.create_time = datetime.now()


class UploadFileManager:
    def __init__(self, storage_path):
        self.upload_file_records = {}
        self.storage_path = storage_path
        self._create_storage_folder()

    def _create_storage_folder(self):
        if os.path.exists(self.storage_path):
            shutil.rmtree(self.storage_path)
        os.makedirs(self.storage_path)

    def store_uploaded_file(self, key, file):
        file_name = key
        file_path = os.path.join(self.storage_path, file_name)
        # save the record
        file_record = UploadFileRecord(file_path)
        self.upload_file_records[key] = file_record

        # save the file to the storage, remove the old one first
        if os.path.exists(file_path):
            os.remove(file_path)

        file.save(file_path)
        # print ("Saved to path: ", file_path)

    def fetch_uploaded_file_path(self, key) -> Tuple[bool, str]:
        """
        Returns a tuple. [status: bool, file_path: str]
        """
        try:
            record = self.upload_file_records[key]
        except:
            return False, "Record doesn't exits"

        try:
            file_path = record.file_path
        except:
            return False, "Bad record"

        return True, file_path



"""
Cosmic-CoNN model training options.
CY Xu (cxu@ucsb.edu)
"""

import argparse
import os
import datetime

class ModelOptions():
    def __init__(self):
        # init default model options
        
        # training and model settings
        self.__dict__["data"] = "./"
        self.__dict__["max_train_size"] = 0
        self.__dict__["mode"] = "inference"
        self.__dict__["model"] = "lco"
        self.__dict__["loss"] = 0
        self.__dict__["imbalance_alpha"] = 100.0
        self.__dict__["norm"] = "group"
        self.__dict__["no_affine"] = False
        self.__dict__["n_group"] = 8
        self.__dict__["gn_channel"] = 0
        self.__dict__["conv_type"] = "unet"
        self.__dict__["up_type"] = "deconv"
        self.__dict__["down_type"] = "maxpool"
        self.__dict__["deeper"] = False
        self.__dict__["crop"] = 1024
        self.__dict__["batch"] = 4
        self.__dict__["hidden"] = 32
        self.__dict__["lr"] = 0.001
        self.__dict__["eval_epoch"] = 10
        self.__dict__["min_exposure"] = 0.0
        self.__dict__["epoch"] = 3000
        self.__dict__["milestones"] = ['0']
        self.__dict__["seed"] = 0
        self.__dict__["random_seed"] = False
        self.__dict__["comment"] = ""

        # validation/test settings
        self.__dict__["load_model"] = ""
        self.__dict__["validate_freq"] = 1
        self.__dict__["validRatio"] = 0.2
        self.__dict__["max_valid_size"] = 0
        self.__dict__["valid_crop"] = 2000

        # continue training
        self.__dict__["continue_train"] = ""
        self.__dict__["continue_epoch"] = 0

        self.__dict__["expr_dir"] = "./"
        self.__dict__["log_file"] = ""
        self.__dict__["prefix"] = ""

        # inference settings
        self.__dict__["app"] = False
        self.__dict__["input"] = "./"
        self.__dict__["ext"] = "SCI"


    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __len__(self):
        return len(self.__dict__)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()


class TrainingOptions:
    def __init__(self):
        self.parser = argparse.ArgumentParser()
        self.initialized = False

    def initialize(self):
        # training and model settings
        self.parser.add_argument(
            "--data",
            default=os.path.join(os.getcwd(), "data"),
            help="path to data directory",
        )

        self.parser.add_argument(
            "--max_train_size",
            type=int,
            default=0,
            help="the # of samples randomly draw in each epoch, 0 uses entire dataset",
        )
        self.parser.add_argument(
            "--mode",
            type=str,
            default="train",
            help="train | inference, inference mode does not create checkpoint or log",
        )
        self.parser.add_argument(
            "--model",
            type=str,
            default="lco",
            help="lco | hst | nres, dataset specific dataloader",
        )
        self.parser.add_argument(
            "--loss",
            type=str,
            default="median_bce",
            help="bce | median_bce | dice | mse, loss function used for training",
        )
        self.parser.add_argument(
            "--imbalance_alpha",
            type=float,
            default=100.0,
            help="number of iterations for the Median Weighted BCE Loss to linearly increase the lower bound alpha to 1. See paper for detail.",
        )
        self.parser.add_argument(
            "--norm",
            type=str,
            default="group",
            help="batch | group | instance, feature normalization method",
        )
        self.parser.add_argument(
            "--no_affine",
            action="store_true",
            help="do not use affine transformation in normalization",
        )
        self.parser.add_argument(
            "--n_group", type=int, default=8, help="fixed group number for group normalization "
        )
        self.parser.add_argument(
            "--gn_channel",
            type=int,
            default=0,
            help="fixed channel number, >0 will override n_group, 0 uses fixed group number",
        )
        self.parser.add_argument(
            "--conv_type", type=str, default="unet", help="unet | resnet, types for convolution module"
        )
        self.parser.add_argument(
            "--up_type", type=str, default="deconv", help="deconv | upscale, types for deconvolution module"
        )
        self.parser.add_argument(
            "--down_type",
            type=str,
            default="maxpool",
            help="maxpool | avgpool | stride, types for the pooling layer",
        )
        self.parser.add_argument(
            "--deeper",
            action="store_true",
            help="deeper network, one more downsample and upsample layer",
        )
        self.parser.add_argument(
            "--crop", type=int, default=1024, help="training input stamp size"
        )
        self.parser.add_argument(
            "--batch",
            type=int,
            default=4,
            help="training batch size",
        )
        self.parser.add_argument(
            "--hidden", type=int, default=32, help="channel # of first conv layer"
        )
        self.parser.add_argument(
            "--lr", type=float, default=0.001, help="learning rate, 0.001 by default"
        )
        self.parser.add_argument(
            "--eval_epoch", type=int, default=10, help="# of phase0 training epochs, only applies to batch normalization"
        )
        self.parser.add_argument(
            "--min_exposure",
            type=float,
            default=0.0,
            help="minimum exposure time when sampling training data",
        )

        self.parser.add_argument(
            "--epoch", type=int, default=3000, help="total training epochs")
        self.parser.add_argument(
            "--milestones",
            nargs="+",
            default=['0'],
            help="milesstones to reduce the learning rate, e.g. '1000 2000 3000', '0' keeps LR constant",
        )
        self.parser.add_argument(
            "--seed", type=int, default=0, help="assign manual seed")
        self.parser.add_argument(
            "--random_seed", action="store_true", help="it will initialize the model multiple times to find a best random seed if flagged"
        )
        self.parser.add_argument(
            "--comment", default="", help="comment is appended to the checkpoint directory name")

        # validation/test settings
        self.parser.add_argument(
            "--load_model", default="", help="path to load a model for inference model")
        self.parser.add_argument(
            "--validate_freq", type=int, default=1, help="number per epochs to perform model validation")
        self.parser.add_argument(
            "--validRatio", type=float, default=0.2, help="the ratio of training data reserved for validation, 0.2 by default")
        self.parser.add_argument(
            "--max_valid_size", type=int, default=0, help="the number of sample reserved for validation, >0 will override validRatio")
        self.parser.add_argument(
            "--valid_crop", type=int, default=2000, help="stamp size for the center-cropping during validation")

        # continue training
        self.parser.add_argument(
            "--continue_train", type=str, default="", help="to continue a previous training, provide the checkpoint directory name"
        )
        self.parser.add_argument(
            "--continue_epoch", type=int, default=0, help="the number of epoch to continue")

        self.initialized = True
        self.opt = self.parser.parse_args()

    def parse(self):
        if not self.initialized:
            self.initialize()

        self.opt.expr_dir, self.opt.prefix = self.mkCheckpointDir(self.opt)

        args = vars(self.opt)

        if self.opt.mode == "train":

            print("------------ Options -------------")
            for k, v in args.items():
                print("%s: %s" % (str(k), str(v)))
            print("-------------- End ----------------")

            if self.opt.continue_train:
                file_name = os.path.join(
                    self.opt.expr_dir,
                    f"opt_train_continue_{self.opt.continue_epoch}.txt",
                )

            if not (self.opt.continue_train or self.opt.load_model):
                # save to the disk
                file_name = os.path.join(self.opt.expr_dir, "opt_train.txt")

            with open(file_name, "wt") as opt_file:
                opt_file.write("------------ Options -------------\n")
                for k, v in args.items():
                    opt_file.write("%s: %s\n" % (str(k), str(v)))
                opt_file.write("-------------- End ----------------\n")

            self.opt.log_file = file_name

        # swtich to custom class from argparse dict
        model_options = ModelOptions()
        for k, v in args.items():
            model_options[k] = v

        return model_options


    def mkCheckpointDir(self, opt):
        suffix = opt.comment
        continue_train = opt.continue_train

        if continue_train:
            directory = continue_train
            path = f"./checkpoints/{continue_train}"

        else:
            now = datetime.datetime.now()
            timestamp = now.strftime("%Y_%m_%d_%H_%M")
            directory = str(timestamp) + "_" + suffix
            path = f"./checkpoints/{directory}"

            if self.opt.mode == "train":
                os.makedirs(path, exist_ok=True)

        return path, directory



# -*- coding: utf-8 -*-

"""
inference_cr.py: main entry file for console commands, app, and package import.
See documentation for usage.
CY Xu (cxu@ucsb.edu)
"""

import os
import time
import shutil
from tqdm import tqdm
import pretty_errors

import torch.backends.cudnn as cudnn
from astropy.io import fits

from cosmic_conn.dl_framework.cosmic_conn import Cosmic_CoNN
from cosmic_conn.dl_framework.options import ModelOptions
from cosmic_conn.data_utils import check_trained_models, console_arguments
from cosmic_conn.data_utils import parse_input

cudnn.enabled = True
cudnn.benchmark = True

PREDICT_DIR = "cosmic_conn_output"
TEMP_DIR = "instance_temp_storage"
MODEL_VERSON = "0.2.2"


def init_model(model, opt=None):
    """Models are initialized here by passing in the keyword.
    return a pytorch model.
    """
    print("Initializing Cosmic-CoNN CR detection model...")

    model_dir = check_trained_models()

    if opt is None:
        opt = ModelOptions()

    # overwrite model if init_model is called directly
    opt.model = model

    # model path
    if opt.model == "ground_imaging":
        opt.load_model = f"{model_dir}/cosmic-conn_ground_imaging_v{MODEL_VERSON}.pth"

    elif opt.model == "NRES":
        opt.load_model = f"{model_dir}/cosmic-conn_LCO_NRES_v{MODEL_VERSON}.pth"

    elif opt.model == "HST_ACS_WFC":
        opt.norm = "batch"
        opt.load_model = f"{model_dir}/cosmic-conn_HST_ACS_WFC_v{MODEL_VERSON}.pth"

    elif opt.model.endswith("tar") or opt.model.endswith("pth"):
        opt.load_model = opt.model

    else:
        raise ValueError(
            "-m [ground_imaging | NRES | HST_ACS_WFC | model_file(.pth/.tar)]"
        )

    # init model instance
    cr_model = Cosmic_CoNN()
    cr_model.initialize(opt)
    cr_model.eval()

    print(f"{opt.model} model initialized.\n")
    return cr_model


def detect_FITS(model):
    # get arguments from user
    options = model.opt
    all_fits = parse_input(options.input, PREDICT_DIR)
    ext = model.opt.ext

    # start batch detection
    for i in tqdm(range(len(all_fits))):
        f = all_fits[i]

        with fits.open(f) as hdul:
            image = None
            tic = time.perf_counter()

            try:
                if ext != 'SCI':
                    print(f"Reading data from hdul[{ext}]")
                    # if user asigned extension, try it first
                    image = hdul[ext].data.astype("float32")

                elif hdul[0].data is not None:
                    print("Reading data from hdul[0], "
                          "to specify extension name: -e SCI.")
                    image = hdul[0].data.astype("float32")

                elif hdul[1].data is not None:
                    print("Reading data from hdul[1], "
                          "to specify extension name: -e SCI.")
                    image = hdul[1].data.astype("float32")

                else:
                    print(f"Reading data from hdul[{ext}], "
                          "to specify extension name: -e SCI.")
                    image = hdul[ext].data.astype("float32")

            except:
                raise ValueError(
                    f"No valid data found in extention 0, 1 or {ext}, "
                    "to specify extension name: -e SCI.")

            # detection
            cr_probability = model.detect_cr(image, ret_numpy=True)

            hdul.append(fits.CompImageHDU(
                cr_probability, name="CR_probability"))

            # a new copy is saved in the prediction directory
            out_dir = os.path.join(os.path.split(f)[0], PREDICT_DIR)
            os.makedirs(out_dir, exist_ok=True)

            out_name = os.path.join(out_dir, f"CR_{os.path.basename(f)}")
            hdul.writeto(out_name, overwrite=True)

            toc = time.perf_counter()
            print(f"Detection of a {image.shape} image took"
                  f"{round(toc-tic, 2)}s.\n"
                  f"Result saved as {os.path.basename(out_name)}\n")

    print(
        f"\nDone. The CR probability map is appended to a new "
        f"FITS copy and saved in {PREDICT_DIR}")

    if os.path.exists(TEMP_DIR):
        shutil.rmtree(TEMP_DIR)

    return 0


def detect_image(cr_model, image, ret_numpy=True):
    cr_mask = cr_model.detect_cr(image.astype("float32"), ret_numpy=ret_numpy)
    return cr_mask


def CLI_entry_point():
    # entry point form CLI detection or web app
    opt = console_arguments()
    cr_model = init_model(opt.model, opt)

    if opt.app:
        # launch web-app
        from cosmic_conn.web_app import app        
        app.main(cr_model, opt)
    else:
        # batch processing FITS
        detect_FITS(cr_model)



# -*- coding: utf-8 -*-

"""
Help functions for data IO
CY Xu (cxu@ucsb.edu)
"""

import os
import numpy as np

from astropy.io import fits
from reproject import reproject_interp

from cosmic_conn.cr_pipeline.utils_img import acquire_valid_area, save_as_png, zscale_stack_cutoff

EXTENSIONS = ["fits", "fz"]


def is_fits_file(filename):
    filename = filename.lower()
    return any(filename.endswith(extension) for extension in EXTENSIONS)


def pop_from_list(list, index):
    new_list = []
    for i in range(len(list)):
        if i not in index:
            new_list.append(list[i])
    return new_list


def group_nres_sequences(root, fnames):
    fnames.sort()
    grouped_sequences = []

    # otherwise group frames of the same requst
    sequence = []

    i = 0
    while i < len(fnames):
        i_path = os.path.join(root, fnames[i])

        try:
            with fits.open(i_path) as hdu:
                hdr = hdu["SPECTRUM"].header

                sequence_propid = hdr["PROPID"]
                sequence_exptime = hdr["EXPTIME"]
                sequence_request = hdr["REQNUM"]

                date = hdr["DATE"]
                object_name = hdr["OBJECT"] if hdr["OBJECT"] != "N/A" else "NA"
        except:
            # os.remove(i_path)
            print(f"{i_path} is corrupted and removed")
            i += 1
            continue

        # inner loop of consecutive frames
        end_idx = min(len(fnames), i + 3)
        sequence = []
        bkgs = []

        for j in range(i, end_idx):

            # read current frame and header
            curr_fname = fnames[j]
            path = os.path.join(root, curr_fname)
            print(curr_fname)

            with fits.open(path) as hdu:
                hdr = hdu["SPECTRUM"].header

                proposal_id = hdr["PROPID"]
                request = hdr["REQNUM"]
                exptime = hdr["EXPTIME"]

            # exposure time are not always exact the same between frames
            # 5% difference allowed but avoid large difference in exposure time
            if np.abs(exptime - sequence_exptime) > 0.05 * sequence_exptime:
                print(f"exposure time not matching\n")
                continue

            if request == "UNSPECIFIED":
                print(f"request unspecified\n")
                continue

            # check if following frames are from the same request (in case of missing files)
            if request != sequence_request:
                print(f"request not matching\n")
                continue

            if proposal_id != sequence_propid:
                print(f"proposal id not matching\n")
                continue

            # evertthing checks out, add to list
            sequence.append(curr_fname)

        # dump the sequence to the main list after inner loop
        if len(sequence) == 3:
            grouped_sequences.append(sequence)

            print(
                f"request {sequence_request} took {len(sequence)} frames of {object_name}"
            )
            print(f"on {date[:10]}, exp_time {sequence_exptime}\n")

            # hdus are the "clean" single frames replaced pixels from BPM
            aligned_hdul = new_hdul()

            for f in sequence:
                f_path = os.path.join(root, f)
                with fits.open(f_path) as hdu:
                    extension = hdu["SPECTRUM"]
                    extension.header["maxlin"] = 30000
                    extension.header["saturate"] = 37500
                    extension.header["filename"] = f

                    aligned_hdul.append(extension.copy())

            # update primary header
            aligned_hdul[0].header["frmtotal"] = len(sequence)

            # save stacked frames in one fits
            root_parent = os.path.split(root)[0]
            # stacked_root = os.path.join(root_parent, 'aligned_nres')
            save_as_fits(aligned_hdul, sequence[0], root_parent)

        # move forward frame_count steps in the outer loop
        # avoid infinite loop, skip ahead one frame
        if len(sequence) == 0:
            i += 1
        else:
            i += len(sequence)

        if i >= len(fnames):
            break

    return grouped_sequences


def group_consecutive_sequences(root, fnames, min_exptime):
    fnames.sort()
    grouped_sequences = []

    # otherwise group frames of the same requst
    sequence = []

    i = 0
    while i < len(fnames):
        i_path = os.path.join(root, fnames[i])

        try:
            with fits.open(i_path) as hdu:
                hdr = hdu["SCI"].header

                sequence_frmtotal = hdr["FRMTOTAL"]
                sequence_exptime = hdr["EXPTIME"]
                sequence_request = hdr["REQNUM"]
                sequence_molecule = hdr["MOLNUM"]

                date = hdr["DATE"]
                object_name = hdr["OBJECT"] if hdr["OBJECT"] != "N/A" else "NA"

                # some fits misses WCSERR key
                try:
                    wcserr = hdr["WCSERR"]
                except:
                    wcserr = 1
        except:
            os.remove(i_path)
            print(f"{i_path} is corrupted and removed")
            i += 1
            continue

        # keep only consecutive sequence
        if sequence_frmtotal < 3 or sequence_frmtotal > 12:
            os.remove(i_path)
            print(f"{fnames[i]} removed, frame count {sequence_frmtotal}\n")
            i += 1
            continue

        # reject short-exposed frames
        elif sequence_exptime < min_exptime:
            i += 1
            print(f"exposure less than {min_exptime}")
            continue

        # reject WCS unresolved frames
        elif wcserr != 0:
            os.remove(i_path)
            i += 1
            print("wcs not resolved")
            continue

        else:
            # 0m4 telescope shutter tend to open before telescope stablize
            # leading to blurry first exposure, skip 1st frame if possible
            if sequence_frmtotal > 3 and fnames[i][3:6] == "0m4":
                i += 1

            # inner loop of consecutive frames
            end_idx = min(len(fnames), i + 3)
            sequence = []
            bkgs = []

            for j in range(i, end_idx):

                # read current frame and header
                curr_fname = fnames[j]
                path = os.path.join(root, curr_fname)
                print(curr_fname)

                with fits.open(path) as hdu:
                    hdr = hdu["SCI"].header
                    request = hdr["REQNUM"]
                    exptime = hdr["EXPTIME"]
                    molecule = hdr["MOLNUM"]

                    try:
                        bkgs.append(float(hdr["L1MEDIAN"]))
                    except:
                        os.remove(path)
                        fnames.pop(j)
                        print(f"L1MEDIAN not available\n")
                        continue

                # exposure time are not always exact the same between frames
                # 5% difference allowed but avoid large difference in exposure time
                if np.abs(exptime - sequence_exptime) > 0.05 * sequence_exptime:
                    print(f"exposure time not matching\n")
                    continue

                # check if following frames are from the same request (in case of missing files)
                if request != sequence_request:
                    print(f"request not matching\n")
                    continue

                # same request could split into different molecules, should avoid
                if molecule != sequence_molecule:
                    print(
                        f"molecule {molecule}, sequence molecule {sequence_molecule}\n"
                    )
                    continue

                # evertthing checks out, add to list
                sequence.append(curr_fname)

                # dump the sequence to the main list after inner loop
                if len(sequence) == 3:

                    # test sample standard deviation between frames
                    # s = sqrt((1/(N-1)sum(x_i-mu)^2))
                    # for numpy, ddof makes it unbiased std

                    bkgs = np.array(bkgs) - np.median(bkgs)
                    std = np.std(bkgs, ddof=1)

                    if std < 5.0:
                        grouped_sequences.append(sequence)

                        print(
                            f"request {round(float(sequence_request))}",
                            f"took {len(sequence)} frames of {object_name}",
                            f"on {date[:10]}, exp_time {sequence_exptime},",
                            f"molecule {sequence_molecule}\n",
                        )

                        # reset after 3 frames
                        # sequence_copy = copy.deepcopy(sequence)
                    else:
                        print(curr_fname)
                        print(f"bkg variation {std} too too high\n")

            # move forward frame_count steps in the outer loop
            # avoid infinite loop, skip ahead one frame
            if len(sequence) == 0:
                i += 1
            else:
                i += len(sequence)

        if i >= len(fnames):
            break

    return grouped_sequences


def extract_fname(file_path):
    filename = os.path.split(file_path)[-1]
    if filename.endswith(".fz"):
        return filename[:-8]
    elif filename.endswith(".fitz"):
        return filename[:-5]
    else:
        return os.path.splitext(filename)[0]


def new_hdul():
    hdr = fits.Header()
    hdr["aligned"] = False
    hdr["frmtotal"] = 0
    empty_primary = fits.PrimaryHDU(header=hdr)
    return fits.HDUList([empty_primary])


def save_file_name_to_header(hdul, fpath):
    filename = extract_fname(fpath)
    hdul["SCI"].header["filename"] = filename
    return hdul


def fix_bad_ccd_saturated_pixels(hdul):
    mask = hdul["BPM"].data
    # bad pixels
    median = np.median(hdul["SCI"].data)
    hdul["SCI"].data[mask == 1] = median
    hdul["SCI"].data[mask == 3] = median

    # over-saturated pixels
    saturate = hdul["SCI"].header["saturate"]
    upper_bound = 0.9 * saturate
    hdul["SCI"].data[mask == 2] = upper_bound

    return hdul


def remove_nan(frame):
    # replace Nan value with 0
    nan_index = np.isnan(frame)
    frame[nan_index] = 0.0
    return frame


def copy_WCS(reference_hdu, query_hdu):
    # after projection, replace WCS to the target WCS
    ref_hdr = reference_hdu.header
    keys = ["CRVAL1", "CRVAL2", "CRPIX1", "CRPIX2",
            "CD1_1", "CD1_2", "CD2_1", "CD2_2"]

    for k in keys:
        query_hdu.header[k] = ref_hdr[k]

    return query_hdu


def reproject_to_same_WCS(fits_paths, root, no_png=False):
    """
    Takes N consecutive frames as input, project to middle frame's WCS
    remove bad pixels from banzai's mask, combine to a single new fits  
    return a combined fits, where masked_fits[1:N+1] are reprojected frames 
    """
    # Use middle image as the target for alignment
    N_frames = len(fits_paths)
    reference_idx = N_frames // 2

    # hdus are the "clean" single frames replaced pixels from BPM
    hdus = []

    for i in range(len(fits_paths)):

        path = fits_paths[i]
        # pre-process and copy the frame
        hdul = fits.open(path, lazy_load_hdus=False)
        hdul = fix_bad_ccd_saturated_pixels(hdul)
        hdul = save_file_name_to_header(hdul, path)
        hdus.append(hdul["SCI"].copy())

        # save CAT extension, the extracted sources
        if i == reference_idx:
            cat_extension = hdul["CAT"].copy()

        hdul.close()

    reference_hdu = hdus[reference_idx]
    aligned_hdul = new_hdul()
    valid_masks = []

    for i in range(N_frames):
        hdus[i].ver = i
        query_hdu = hdus[i]

        if i == reference_idx:
            # append the reference frame directly, no reprojection needed
            aligned_hdul.append(query_hdu)

        else:
            # project query_hdu to WCS of the reference_hdu (the middle frame)
            aligned_array, footprint = reproject_interp(
                query_hdu, reference_hdu.header, order="nearest-neighbor"
            )

            aligned_array = remove_nan(aligned_array)
            query_hdu.data = np.float32(aligned_array)

            query_hdu = copy_WCS(reference_hdu, query_hdu)
            aligned_hdul.append(query_hdu)

            # save footprint mask for future use
            valid_masks.append(np.array(footprint > 0).astype("uint8"))

    # from all reprojection footprint masks, get common valid area across all frames
    valid_mask = acquire_valid_area(np.stack(valid_masks))

    aligned_hdul.append(fits.CompImageHDU(
        valid_mask, name="valid_mask", uint=True))

    # append CAT extention, SEP extracted sources
    aligned_hdul.append(cat_extension)

    # update primary header
    aligned_hdul[0].header["frmtotal"] = N_frames

    # save png preview to quickly review/remove bad frames
    if not no_png:
        contrast = 0.1
        hdu_vis = np.stack([hdu.data for hdu in aligned_hdul[1:-1]])
        hdu_vis, _ = zscale_stack_cutoff(
            hdu_vis, nsamples=10_000, contrast=contrast)

        out_path = os.path.join(os.path.split(root)[0], "aligned_png")

        for i in range(N_frames):
            fname = extract_fname(fits_paths[i])
            save_as_png(hdu_vis[i], out_path, fname, file_type="noisy")

    return aligned_hdul


def align_banzai(opt):
    """
    Read banzai processed frames, reproject a consecutive sequence of frames 
    to the same WCS, and combine multiple fits into a single fits

    expected directory structure:
        data/banzai_frms
        data/telescope_a/banzai_frms
        data/telescope_b/banzai_frms

    new fits will be saved to 'aligned_fits' directory in next to the source banzai_frms
    """
    for root, dirs, files in os.walk(opt.data):

        # avoid temp files generated by Mac's Finder
        files = [f for f in files if not f[0] == "."]
        dirs[:] = [d for d in dirs if not d[0] == "."]

        fitz_only = [fit for fit in files if is_fits_file(fit)]
        fitz_only.sort()

        if not root.endswith("banzai_frms"):
            continue
        print(f"{len(fitz_only)} fits files found in {root}\n")

        # group consecutive observations of same object from the same night
        if opt.nres:
            grouped_sequences = group_nres_sequences(root, fitz_only)

        else:
            grouped_sequences = group_consecutive_sequences(
                root, fitz_only, opt.min_exptime
            )

            for i in range(len(grouped_sequences)):

                # each group is N (N>=3) frames of consecutive observations
                # reproject all frames to align with the middle frame's WCS
                fits_paths = [os.path.join(root, f)
                              for f in grouped_sequences[i]]

                aligned_hdul = reproject_to_same_WCS(
                    fits_paths, root, opt.no_png)

                # save aligned frames in one fits
                root_parent = os.path.split(root)[0]
                save_as_fits(
                    aligned_hdul, grouped_sequences[i][0], root_parent)


def save_as_fits(aligned_fits, fname, dir, mask=False):
    # new file name same as the first file of the sequence
    # currently fixed to merge three consecutive frames
    frame_count = 3

    if mask:
        new_path = os.path.join(dir, "masked_fits")
        new_filename = f"{remove_ext(fname)}_{frame_count}frms_masks.fits"
    else:
        new_path = os.path.join(dir, "aligned_fits")
        new_filename = f"{remove_ext(fname)}_{frame_count}frms_aligned.fits"

    # save aligned fits to sub-directory
    mkdir(new_path)
    out_path = os.path.join(new_path, new_filename)

    aligned_fits.writeto(out_path, overwrite=True)


def read_aligned_fits(directory):
    file_names, roots = [], [], []

    for root, dirs, files in os.walk(directory):
        if root.endswith("aligned_fits"):
            # avoid temp files generated by Mac's Finder
            file_names = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]

            file_names.sort()
            roots.append(root)

    return roots, file_names


def remove_outlier_frm(frames, outlier_idx, update_hdul=False):
    if update_hdul:
        # for hdul update, need to remove frame and valid mask
        total = frames[0].header["frmtotal"]

        # hdul[0] is primary frame, skip
        hdul_idx = [x + 1 for x in outlier_idx]
        frames = np.delete(frames, hdul_idx, axis=0)

        # update the count of total frames in headers
        for i in range(len(frames)):
            frames[i].header["frmtotal"] = total - len(outlier_idx)

        frames = fits.HDUList(list(frames))

    else:
        frames = np.delete(frames, outlier_idx, axis=0)

    return frames


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def remove_ext(filename):
    if filename.endswith(".fz"):
        return filename[:-8]
    elif filename.endswith(".fitz"):
        return filename[:-5]
    else:
        return os.path.splitext(filename)[0]


def save_fits(array, dir, filename, file_type):
    fname = filename + file_type + ".fitz"
    out_path = os.path.join(dir, fname)
    hud = fits.PrimaryHDU(array)
    hud.writeto(out_path)
    print(f"{file_type} saved to {out_path}")


def save_as_npy(array, dir, filename, file_type):
    mkdir(dir)
    path = os.path.join(dir, filename + "_" + file_type)
    if file_type == "mask":
        array = array.astype("uint8")
    np.save(path, array, allow_pickle=True)


def save_array_png(
    noisy, mask, cleaned, ignore_mask, root, fnames, save_np, save_png, header, comment
):
    # save frames as numpy arrays
    if save_np:
        npy_dir = os.path.join(root, "train_npy")
        mkdir(npy_dir)

        filename = os.path.join(npy_dir, fnames[0])
        np.savez_compressed(filename, noisy=noisy, mask=mask)

    if save_png:
        png_dir = comment if comment is not None else "train_set_png"
        png_dir = "rejected_frms" if "reject" in comment else png_dir

        img_dir = os.path.join(root, png_dir)

        obj = header["OBJECT"]
        exptime = header["EXPTIME"]
        rdnoise = header["RDNOISE"]
        put_text = f"object:{obj} / exposure:{exptime} / RD_noise:{rdnoise}"

        noisy_vis, cut_offs = zscale_stack_cutoff(
            noisy, nsamples=10_000, contrast=0.1)

        if cleaned is not None:
            clean_vis, _ = zscale_stack_cutoff(
                cleaned, nsamples=10_000, contrast=0.1, cut_offs=cut_offs
            )

        for i in range(noisy.shape[0]):
            fname = fnames[i] + "_" + \
                comment if comment is not None else fnames[i]

            save_as_png(noisy_vis[i], img_dir, fname, "noisy", zscale=False)

            if mask is not None:
                save_as_png(mask[i], img_dir, fname, "mask")

            if cleaned is not None:
                save_as_png(clean_vis[i], img_dir, fname,
                            "cleaned", zscale=False)

        # save single sources mask
        if ignore_mask is not None:
            for j in range(ignore_mask.shape[0]):
                save_as_png(
                    ignore_mask[j], img_dir, fnames[0] +
                    "_ignore_mask", "ignore_mask",
                )


def save_fits_with_CR(opt, hdul, gains, cr_mask, ignore_mask, filenames, debug_frames):
    N_frames = cr_mask.shape[0]

    if opt.debug:
        mask_high, mask_low, cr_signal_to_noise = debug_frames

    for j in range(N_frames):
        # update frame read_noise with new
        hdul[j + 1].header["gain"] = gains[j].item()

        # append CR mask
        hdul.append(
            fits.CompImageHDU(
                cr_mask[j], header=hdul[j + 1].header, name="CR", ver=j, uint=True
            )
        )

        # append debugging frames
    if opt.debug:
        for j in range(N_frames):
            hdul.append(
                fits.CompImageHDU(
                    mask_high[j],
                    header=hdul[j + 1].header,
                    name="mask_high",
                    ver=j,
                    uint=True,
                )
            )
        for j in range(N_frames):
            hdul.append(
                fits.CompImageHDU(
                    cr_signal_to_noise[j],
                    header=hdul[j + 1].header,
                    name="CR_SNR",
                    ver=j,
                )
            )

    # append source mask
    for j in range(ignore_mask.shape[0]):
        hdul.append(
            fits.CompImageHDU(
                ignore_mask[j], header=hdul[j + 1].header, name="ignore", uint=True
            )
        )

    # save as new fits with masks, inside "masked_fits" directory
    save_as_fits(hdul, filenames[0], opt.root, mask=True)


def hdul_to_array(fits):
    # read frames from hdul to ndarrays
    frmtotal = fits[0].header["frmtotal"]

    frames = np.stack([fits[i].data for i in range(1, frmtotal + 1)])

    # valid mask is the common valid area across all frames
    valid_mask = fits[-2].data
    frames *= valid_mask

    filenames = [fits[i].header["filename"] for i in range(1, frmtotal + 1)]
    gains = [fits[i].header["gain"] for i in range(1, frmtotal + 1)]

    return frames, valid_mask, gains, filenames



"""
Defines the basic network components for U-Net.
"""

import torch
import torch.nn as nn


class double_conv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(type(self), self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch, momentum=0.005),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch, momentum=0.005),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(type(self), self).__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x


class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(type(self), self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool2d(2), double_conv(in_ch, out_ch))

    def forward(self, x):
        x = self.mpconv(x)
        return x


class up(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(type(self), self).__init__()
        self.up = nn.ConvTranspose2d(in_ch, out_ch, 2, stride=2)
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x


class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(type(self), self).__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, 1)

    def forward(self, x):
        x = self.conv(x)
        return x



# -*- coding: utf-8 -*-

"""
Define payload classes for data transmission between front-end and back-end.
Payload Types should be consistent between the backend and the front end
Boning Dong, CY Xu (cxu@ucsb.edu)
"""

import numpy as np


class PayloadTypes:
    NONE_TYPE = 0
    IMAGE_TYPE = 1
    THUMBNAIL_TYPE = 2
    FLOATLIST_TYPE = 3


"""
Payload format:
[payload size: uint32][payload type: uint8][payload]
note: the payload size doesn't include payload size and the payload type field
"""


def add_payload_meta_and_to_bytes(payload):
    payload_bytes: bytes = payload.tobytes()
    payload_size = np.array([len(payload_bytes)], dtype=np.uint32)
    payload_type = np.array([payload.get_payload_type()], dtype=np.uint8)
    payload_bytes = payload_size.tobytes() + payload_type.tobytes() + payload_bytes
    return payload_bytes


"""
PostResponse format
[Payload Count: uint32][ payload 1 ][ payload 2 ]
"""


class PostResponse:
    def __init__(self):
        self.payloads = []

    def append_payload(self, payload):
        self.payloads.append(payload)

    def tobytes(self):
        payload_count = np.array([len(self.payloads)], dtype=np.uint32)
        response_bytes = payload_count.tobytes()
        for payload in self.payloads:
            response_bytes += add_payload_meta_and_to_bytes(payload)
        return response_bytes


"""
ImagePayload format
[width: uint32][height: uint32][ image array: float32 ]
"""


class ImagePayload:
    def __init__(self, img_2d_array):
        h, w = img_2d_array.shape
        self.img_height = h
        self.img_width = w
        self.img_2d_array = img_2d_array

    def get_payload_type(self):
        return PayloadTypes.IMAGE_TYPE

    def tobytes(self):
        img_meta = np.array([self.img_width, self.img_height], dtype=np.uint32)
        img_2d_array_float32 = self.img_2d_array.astype("float32")
        img_payload_bytes = img_meta.tobytes() + img_2d_array_float32.tobytes()
        return img_payload_bytes


"""
ThumbnailPayload format
[thumbnail number: uint32][patch size: uint32][coordinate array: x: uint32, y: uint32, x:uint32, y:uint32 ...]
"""


class ThumbnailPayload:
    def __init__(self, patch_size):
        self.patch_size = patch_size
        self.thumbnail_coords = []

    def add_thumbnail_coord(self, coord):
        self.thumbnail_coords.append(coord)

    def get_payload_type(self):
        return PayloadTypes.THUMBNAIL_TYPE

    def tobytes(self):
        thumbnail_meta = np.array(
            [len(self.thumbnail_coords), self.patch_size], dtype=np.uint32
        )
        thumbnail_body = np.array(
            self.thumbnail_coords).flatten().astype(np.uint32)
        thumbnail_payload_bytes = thumbnail_meta.tobytes() + thumbnail_body.tobytes()
        return thumbnail_payload_bytes


"""
Floating Point List format
[length of list: uint32][list of data in float32: [zscale z1, zscale z2, ...]]
"""


class FloatListPayload:
    def __init__(self, list_of_data):
        # self.length = len(list_of_data)
        self.data = list_of_data

    def get_payload_type(self):
        return PayloadTypes.FLOATLIST_TYPE

    def tobytes(self):
        # length = np.array(self.length, dtype=np.uint32).tobytes()
        zscale = np.array(self.data[:2], dtype="float32").tobytes()
        # length +
        return zscale



"""
Defines dataloader for CR datasets of various types of data.
CY Xu (cxu@ucsb.edu)
"""

import os
import random
import numpy as np

from astropy.io import fits
from astropy.io.fits import getval
from torch.utils.data import Dataset, DataLoader

from cosmic_conn.dl_framework.utils_ml import (
    _worker_init_fn,
    center_crop_npy,
    erase_boundary_np,
    modulus_boundary_crop,
)


def is_fits_file(filename):
    filename = filename.lower()
    return any(filename.endswith(extension) for extension in ["fits", "fz"])


def find_all_fits(opt, min_exposure=100.0):
    # key_word = 'banzai_frms' if opt.load_model else 'masked_fits'
    paths, file_names = [], []
    skipped = 0

    for root, dirs, files in os.walk(opt.data):
        if "masked_fits" in root and len(files) > 0:
            files = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]
            files.sort()
            for f in files:
                p = os.path.join(root, f)
                if is_fits_file(p):
                    try:
                        try:
                            exp_time = getval(p, "EXPTIME", 1)
                        except:
                            # Gemini header
                            exp_time = getval(p, "EXPTIME", 0)
                        if exp_time > min_exposure:
                            paths.append(p)
                            file_names.append(f)
                        else:
                            skipped += 1
                    except:
                        print(f"problem reading fits: {p}")
                        # os.remove(p)

    path_name_pairs = list(zip(paths, file_names))
    print(f"{skipped} fits skipped for exposure time shorter than {min_exposure}")

    return path_name_pairs


# @profile
def CreateDataLoader(opt):

    if opt.model in ["mixed", "lco", "nres"]:
        print(f"loading LCO data...")
        all_samples = find_all_fits(opt, min_exposure=opt.min_exposure)
        assert len(all_samples) > 0, \
            f"No valid training data found in {opt.data}"

    if opt.model in ["mixed", "hst"]:
        print(f"loading HST data...")
        all_samples = read_HST_patches(opt.data)

    if opt.max_valid_size > 0:
        valid_size = \
            min(opt.max_valid_size, int(opt.validRatio * len(all_samples)))
    else:
        valid_size = int(opt.validRatio * len(all_samples))

    train_size = len(all_samples) - valid_size

    # in-place shuffle
    random.shuffle(all_samples)

    train_samples = all_samples[:train_size]
    valid_samples = all_samples[train_size:]

    if valid_size < 1:
        valid_samples = train_samples

    # when creating dataload only for evaluations
    if train_size > 0:

        if opt.model == "lco":
            train_set = LCO_Dataset(opt, train_samples)

        elif opt.model == "nres":
            train_set = LCO_NRES_Dataset(opt, train_samples)

        # created HST and LCO dataest and a fused dataset
        elif opt.model == "hst":
            train_set = HST_dataset(
                train_samples,
                mini_batch=1,
                augment=5,
                max_size=opt.max_train_size,
                mode="train",
            )

        else:
            raise ValueError(f"uknown training model {opt.model}")

        # create training dataloader
        train_data_loader = DataLoader(
            dataset=train_set,
            batch_size=opt.batch,
            shuffle=True,
            drop_last=True,
            num_workers=8,
            pin_memory=True,
            worker_init_fn=_worker_init_fn,
        )

    else:
        train_data_loader = None

    # Create validation set and data loader
    if opt.model == "lco":
        valid_set = LCO_Dataset(opt, valid_samples, validate=True)

    elif opt.model == "hst":
        valid_set = HST_dataset(
            valid_samples,
            mini_batch=64,
            augment=None,
            max_size=opt.max_valid_size,
            mode="validate",
        )

    elif opt.model == "nres":
        valid_set = LCO_NRES_Dataset(opt, valid_samples, validate=True)

    else:
        raise ValueError(f"uknown training model {opt.model}")

    if valid_set:
        valid_batch = 16 if opt.model == "hst" else 1

        valid_data_loader = DataLoader(
            dataset=valid_set,
            batch_size=valid_batch,
            shuffle=False,
            drop_last=False,
            num_workers=4,
            pin_memory=True,
        )

    return train_data_loader, train_set, valid_data_loader


def read_HST_patches(path):
    all_paths = []

    for root, dirs, files in os.walk(path):

        files = [f for f in files if not f[0] == "."]
        files = [f for f in files if f.endswith("npz")]
        files.sort()

        for f in files:
            all_paths.append(os.path.join(root, f))

    assert len(all_paths) > 0, f"no HST image found, check {path}"

    return all_paths


# modified by CY
def identical_random_crop(array, size, crop_indices=None):
    h, w = array.shape
    rangew = (w - size) // 2 if w > size else 0
    rangeh = (h - size) // 2 if h > size else 0

    if crop_indices is None:
        offsetw = np.random.randint(0, rangew)
        offseth = np.random.randint(0, rangeh)
    else:
        offsetw = crop_indices[0]
        offseth = crop_indices[1]

    offsetw = 0 if rangew == 0 else offsetw
    offseth = 0 if rangeh == 0 else offseth

    cropped = array[offseth: offseth + size, offsetw: offsetw + size]

    if cropped.shape[-1] == 0:
        return array, [offsetw, offseth]
    else:
        return cropped, [offsetw, offseth]


def stack_identical_random_crop(
    ndarrays, crop_size, crop_indices=None, rotation=None, flip=None
):

    # if no crop_indices provided, create new from mask
    if crop_indices is None:
        patch, crop_indices = identical_random_crop(
            ndarrays[0], crop_size, None)

    # batch random crop frames and masks
    identical_crops = []

    for i in range(ndarrays.shape[0]):
        identical_crops.append(
            identical_random_crop(ndarrays[i], crop_size, crop_indices)[0]
        )

    # stack the same patch from different exposures
    cropped_patch = np.stack(identical_crops)

    # rotate the same patch of sky for data augmentation
    if rotation is not None:
        cropped_patch = np.rot90(cropped_patch, rotation, axes=(-2, -1))

    # flip for data augmentation
    if flip is not None:
        if flip == 0:
            cropped_patch = np.flipud(cropped_patch)
        if flip == 1:
            cropped_patch = np.fliplr(cropped_patch)

    return cropped_patch, crop_indices


def subtact_sky_remove_extreme(noisy_stack):
    # subtract sky and remove extreme low value from entire stack
    noisy_stack -= np.median(noisy_stack)
    lower_bound = np.mean(noisy_stack) - 3.5 * np.std(noisy_stack)
    noisy_stack[noisy_stack < lower_bound] = lower_bound
    return noisy_stack


def sample_cr_frames(frames, cr_frames, min_layers=0, max_layers=9):
    """
    return a randomly cropped CR frames overlay same shaep as frames
    """
    cr_overlay = np.zeros_like(frames)

    if max_layers == 0:
        return cr_overlay

    shape = frames.shape

    if len(shape) == 4:
        n_frames = shape[1]
        crop_shape = shape[-1]
    elif len(shape) == 3:
        n_frames = shape[0]
        crop_shape = shape[-1]
    else:
        raise ValueError("check input frames shape")

    for i in range(n_frames):
        # superimpose random layers of CR for each image
        n_layers = np.random.randint(min_layers, max_layers)
        cr_layers = np.random.choice(cr_frames, n_layers, replace=False)

        curr_overlay = np.zeros([crop_shape, crop_shape], dtype="float32")

        for j in range(n_layers):
            cr_layer = np.load(cr_layers[j], allow_pickle=True)
            # faster array operation without memory map...
            cr_layer = cr_layer["arr_0"]

            cr_crop, _ = identical_random_crop(cr_layer, crop_shape)

            # random rotate for data augmentation
            r = random.randint(0, 3)
            cr_crop = np.rot90(cr_crop, r, axes=(-2, -1))

            # random flip for data augmentation
            f = random.randint(0, 2)
            if f == 0:
                cr_crop = np.flipud(cr_crop)
            if f == 1:
                cr_crop = np.fliplr(cr_crop)

            curr_overlay += cr_crop

        if len(shape) == 4:
            cr_overlay[0, i] = curr_overlay

        if len(shape) == 3:
            cr_overlay[i] = curr_overlay

    return cr_overlay


def superimpose_cr(
    frames, cr_masks, cr_frames, min_layers=0, max_layers=9, source_mask=None
):
    """
    Superimpsoe frames and CR masks with identical cr_frames crops
    """
    cr_overlay = sample_cr_frames(frames, cr_frames, min_layers, max_layers)

    if source_mask is not None:
        cr_overlay = cr_overlay * source_mask

    frames += cr_overlay
    cr_masks = (cr_masks + cr_overlay > 0).astype("uint8")

    return frames, cr_masks


# LCO data helper function


def maximum_center_crop(frame_stack, crop_size):
    """
    crop non-overlapping patches from full frame for validation
    boundary removed for improve robustness
    :return:
    """
    assert (
        len(frame_stack.shape) == 3
    ), f"expecting input as a stack of 2D arrays [n, h, w]"

    n, h, w = frame_stack.shape
    cropped_patches = []

    if crop_size == 0:
        if h % 16 != 0 or w % 16 != 0:

            for i in range(n):
                cropped_patches.append(
                    modulus_boundary_crop(frame_stack[i], modulo=16)
                )

            return np.stack(cropped_patches)

        else:
            return frame_stack

    else:
        if h < crop_size or w < crop_size:
            crop_size = min(h, w)

        for i in range(n):
            cropped_patches.append(
                center_crop_npy(frame_stack[i], crop_size=crop_size)
            )

        return np.stack(cropped_patches)


def sample_frames_from_hdul(key, full_set, validate, opt):
    """
    LCO training data fits extension (example of 3 frames)
    No.         Name                    Format
    0           PRIMARY
    1-3         SCI                     float32
    4           VALID_MASK              uint8
    5           CAT (SEP from banzai)
    6-8         CR_MASK                 uint8
    9-11        SOURCE_MASK             uint8
    """
    file_path = full_set[key]
    hdul = fits.open(file_path, lazy_load_hdus=False)

    # randomly sample 3 frames from a larger stack to save and maintain consistent GPU ram
    frmtotal = hdul[0].header["frmtotal"]

    subset = random.sample(range(frmtotal), 3)

    # skip hdul[0] primary frame
    frames = [hdul[1 + i].data for i in subset]

    # skip primary, valid_mask
    cr_masks = [hdul[3 + frmtotal + i].data for i in subset]

    frames = np.stack(frames, axis=0)
    cr_masks = np.stack(cr_masks, axis=0)

    if validate:
        ignore_masks = [hdul[3 + frmtotal + frmtotal + i].data for i in subset]
        ignore_masks = np.stack(ignore_masks, axis=0)

        headers = [dict(hdul[i + 1].header) for i in subset]
        file_names = ["lco_" + hdr["FILENAME"] for hdr in headers]

    hdul.close()

    # special handling certain telescopes with boundary issues
    # remove 100px around the boundary in image and mask
    edge_width = 100
    frames = erase_boundary_np(frames, edge_width)
    cr_masks = erase_boundary_np(cr_masks, edge_width)

    if validate:
        ignore_masks = erase_boundary_np(ignore_masks, edge_width)

        # center crop for validation data
        frames = maximum_center_crop(frames, opt.valid_crop)
        cr_masks = maximum_center_crop(cr_masks, opt.valid_crop)
        ignore_masks = maximum_center_crop(ignore_masks, opt.valid_crop)

    else:
        # generate identical rotation and flip seeds so both
        # frame and mask are rotated identically
        rotation = random.randint(0, 3)
        flip = random.randint(0, 2)

        # used for deepCR which did not use such augmentation
        if "deepcr" in opt.comment.lower():
            rotation = None
            flip = None

        # only crop training images
        frames, crop_indices = stack_identical_random_crop(
            frames, opt.crop, None, rotation, flip
        )

        cr_masks, _ = stack_identical_random_crop(
            cr_masks, opt.crop, crop_indices, rotation, flip
        )

        file_names = None
        ignore_masks = None

    return frames, cr_masks, ignore_masks, file_names


class LCO_Dataset(Dataset):
    def __init__(self, opt, samples, validate=False):
        self.opt = opt
        self.crop = opt.crop
        self.validate = validate

        # self.subset_paths, self.subset_fnames = zip(*samples)
        self.full_set = {x[1]: x[0] for x in samples}
        self.unused_set = list(self.full_set.keys())

        # self.full_set = samples
        # self.full_set_copy = samples.copy()

        # init and count actual frames
        # self.paths, self.file_names = zip(*self.full_set)
        self.dataset_size = 0

        for item in self.full_set.values():
            with fits.open(item) as hdul:
                self.dataset_size += hdul[0].header["frmtotal"]

        assert self.dataset_size > 0, f"No valid training data found in {self.opt.data}"

        # acquire subset size
        if validate:
            self.subset_size = len(self.full_set)
            print(
                f"Validation set total: {self.dataset_size} frames from {len(samples)} fits\n "
            )
        else:
            if self.opt.max_train_size > 0 and self.opt.max_train_size < len(
                self.full_set
            ):
                print(f"sampling {self.opt.max_train_size} fits each epoch")
                self.subset_size = self.opt.max_train_size
            else:
                print(f"shuffling full data set each epoch")
                self.subset_size = len(self.full_set)

            print(
                f"Training set total: {self.dataset_size} frames from {len(samples)} fits\n "
                f"{opt.crop}x{opt.crop} patch(es) for each fits, batch size {opt.batch} "
            )

        self.sample_subset()

    def __len__(self):
        return self.subset_size

    def sample_subset(self):
        if len(self.unused_set) < self.subset_size:
            self.unused_set = list(self.full_set.keys())

        # sample without duplicates
        self.subset_keys = np.random.choice(
            self.unused_set, self.subset_size, replace=False
        )

        # remove used files to ensure all fiels are used in one epoch
        for s in self.subset_keys:
            if s in self.unused_set:
                self.unused_set.remove(s)

    def __getitem__(self, index):
        # key is the file name
        key = self.subset_keys[index]

        frames, cr_masks, ignore_masks, file_names = sample_frames_from_hdul(
            key, self.full_set, self.validate, self.opt
        )

        cr_masks = cr_masks.astype("float32")
        frames = frames.astype("float32")

        if self.validate:
            ignore_masks = ignore_masks.astype("float32")
        else:
            file_names = "lco_" + key
            ignore_masks = cr_masks

        return frames, cr_masks, ignore_masks, file_names


class HST_dataset(Dataset):
    def __init__(
        self,
        paths=None,
        mini_batch=16,
        augment=None,
        random_nmask=True,
        aug_sky=[0, 0],
        max_size=0,
        mode="train",
    ):
        """Initializa the HSTdata class.

        Parameters:
            file: filename including directory
        """

        # 16x256x256 = 1x1024x1024, so each iteration has equal impact on the model
        self.mini_batch = mini_batch

        self.paths = paths
        self.augment = augment
        self.random_nmask = random_nmask
        self.aug_sky = aug_sky
        self.mode = mode

        # about 60 256x256 images equals one 2k by 2k iamge
        if mode == "evaluate":
            if max_size > 0 and max_size < len(self.paths):
                self.paths = self.paths[:max_size]

        if mode == "train" or mode == "validate":
            if max_size > 0 and max_size < len(self.paths):
                self.paths = random.sample(self.paths, max_size)

        self.dataset_size = len(self.paths)

    def __len__(self):
        return self.dataset_size

    def sample_subset(self):
        return 1

    def __getitem__(self, index):
        # since HST patches are only 256, draw more sample per batch
        if self.mode == "train":
            samples = random.sample(range(self.dataset_size), self.mini_batch)
        else:
            # then not in train mode
            samples = [index]

        raws, cleans, masks, badmasks = [], [], [], []
        gains, filenames = [], []

        for idx in samples:
            path = self.paths[idx]
            patch_npz = np.load(path)

            patch = {}
            for key in ["raw", "clean", "crmask", "badmask", "exp", "gain", "sky"]:
                patch[key] = patch_npz[key]

            exp, gain, sky = patch["exp"], patch["gain"], patch["sky"]

            # generate identical rotation and flip seeds so both
            # frame and mask are rotated identically
            rotation = random.randint(0, 3)
            flip = random.randint(0, 2)

            for key in ["raw", "clean", "crmask", "badmask"]:
                patch[key] = np.rot90(patch[key], rotation, axes=(-2, -1))

                # no flipping if flip == 2
                if flip == 0:
                    patch[key] = np.flipud(patch[key])
                if flip == 1:
                    patch[key] = np.fliplr(patch[key])

            raws.append(patch["raw"].astype("float32"))
            cleans.append(patch["clean"].astype("float32"))
            masks.append(patch["crmask"].astype("float32"))
            badmasks.append(patch["badmask"].astype("float32"))

            filenames.append("hst_" + os.path.basename(path)[:-4])
            gains.append(gain)

        # a = self.aug_sky[0] + np.random.rand() * (self.aug_sky[1] - self.aug_sky[0])
        # a = a * sky

        # if self.augment is not None:
        #     if self.random_nmask:
        #         size = np.random.randint(1, self.augment)
        #     else:
        #         size = self.augment
        #     index = np.random.randint(self.dataset_size, size=size)
        #     index = np.concatenate([index, np.array([i])])
        #     mask = self.crmask[index].sum(axis=0) > 0
        #     badmask = self.badmask[i].astype(bool) + self.crmask[i].astype(bool)

        raws = np.stack(raws)
        cleans = np.stack(cleans)
        masks = np.stack(masks)
        badmasks = np.stack(badmasks)

        return raws, masks, badmasks, cleans, filenames


class LCO_NRES_Dataset(Dataset):
    def __init__(self, opt, samples, validate=False):
        self.opt = opt
        self.crop = opt.crop
        self.full_set = samples
        self.validate = validate

        # init and count actual frames
        self.paths, self.file_names = zip(*samples)

        # count total frames
        self.dataset_size = 0
        for i in range(len(samples)):
            with fits.open(self.paths[i]) as hdul:
                self.dataset_size += hdul[0].header["frmtotal"]

        assert self.dataset_size > 0, f"No valid training data found in {self.opt.data}"

        if validate:
            print(
                f"Validation set: {self.dataset_size} frames from {len(samples)} fits"
            )
        else:
            print(
                f"Training set total: {self.dataset_size} frames from {len(samples)} fits\n "
                f"{opt.crop}x{opt.crop} patch(es) for each fits, batch size {opt.batch}"
            )

        # sample actual epoch subset
        self.sample_subset()

    def __len__(self):
        return len(self.paths)

    def sample_subset(self):
        # shuffle all data each epoch
        subset = random.sample(self.full_set, len(self.full_set))
        self.paths, self.file_names = zip(*subset)

    def sample_nres_frames(self, index):
        """
        LCO NRES training fits extension (example of 3 frames)
        No.         Name                    Format
        0           PRIMARY
        1-3         SCI                     int16
        4-6         CR_MASK                 uint8
        7-9         SOURCE_MASK             uint8
        """
        hdul = fits.open(self.paths[index], lazy_load_hdus=False)

        # randomly sample 3 frames from a larger stack to save and maintain consistent GPU ram
        frmtotal = hdul[0].header["frmtotal"]

        indices = random.sample(range(frmtotal), 3)

        # skip hdul[0] primary frame
        frames = [hdul[1 + i].data for i in indices]
        cr_masks = [hdul[1 + frmtotal + i].data for i in indices]

        # ignore mask is not used during training, only in validation
        if self.validate:
            ignore_masks = [
                hdul[1 + frmtotal + frmtotal + i].data for i in indices
                ]
            ignore_stack = np.stack(ignore_masks, axis=0)
        else:
            ignore_stack = None

        headers = [dict(hdul[i + 1].header) for i in indices]
        file_names = ["nres_" + hdr["FILENAME"] for hdr in headers]

        hdul.close()

        frames = np.stack(frames, axis=0)
        cr_masks = np.stack(cr_masks, axis=0)

        return frames, cr_masks, ignore_stack, file_names

    def __getitem__(self, index):
        frames, cr_masks, ignore_masks, file_names = \
            self.sample_nres_frames(index)

        filename = "nres_" + self.file_names[index]

        # special handling certain telescopes with boundary issues
        # remove 100px around the boundary in image and mask
        edge_width = 100

        frames = erase_boundary_np(frames, edge_width)
        cr_masks = erase_boundary_np(cr_masks, edge_width)

        # No random rotation for NRES data
        rotation = None
        flip = None

        # NRES validation happens on full frame, no cropping
        frames = frames.astype("float32")
        cr_masks = cr_masks.astype("float32")

        # ignore mask is only used during validation
        if self.validate:
            ignore_masks = erase_boundary_np(ignore_masks, edge_width)
            ignore_masks = ignore_masks.astype("float32")

            frames = maximum_center_crop(frames, self.opt.valid_crop)
            cr_masks = maximum_center_crop(cr_masks, self.opt.valid_crop)
            ignore_masks = \
                maximum_center_crop(ignore_masks, self.opt.valid_crop)

        else:
            # only crop training data
            cr_masks, crop_indices = stack_identical_random_crop(
                cr_masks, self.opt.crop, None, rotation, flip
            )

            frames, _ = stack_identical_random_crop(
                frames, self.opt.crop, crop_indices, rotation, flip
            )

            # placeholder, not used
            ignore_masks = cr_masks

        return frames, cr_masks, ignore_masks, filename


class MixedTrainSet(Dataset):
    def __init__(self, opt, LCO_dataset, HST_dataset):
        self.opt = opt
        self.batch_size = opt.batch
        self.batch_counter = 0
        self.die = 0

        self.lco_set = LCO_dataset
        self.hst_set = HST_dataset

        self.lco_len = len(LCO_dataset)
        self.hst_len = len(HST_dataset)

        print(
            f"creating fused dataset, {len(LCO_dataset)} fits from LCO and {len(HST_dataset)} images from HST"
        )

    def __len__(self):
        return self.lco_len

    def __getitem__(self, index):

        if self.batch_counter == 0:
            self.die = random.random()

        elif self.batch_counter == self.batch_size:
            self.batch_counter = 0

        self.batch_counter += 1

        if self.die < 0.5:
            ret = self.lco_set[index]
        else:
            ret = self.hst_set[index]

        return ret


class MixedValidateSet(Dataset):
    def __init__(self, opt, LCO_dataset, HST_dataset):
        self.opt = opt

        self.lco_data = LCO_dataset
        self.hst_data = HST_dataset

        self.lco_len = len(LCO_dataset)
        self.hst_len = len(HST_dataset)

        print(
            f"creating fused validation dataset, {len(LCO_dataset)} fits from LCO and {len(HST_dataset)} images from HST"
        )

    def __len__(self):
        return self.lco_len + self.hst_len

    def __getitem__(self, index):
        # go through both data sets

        if index < self.lco_len:
            return self.lco_data[index]
        else:
            off_set_idx = index - self.lco_len
            return self.hst_data[off_set_idx]



# -*- coding: utf-8 -*-

"""
Helper functions for numpy array and image manipulations.
CY Xu (cxu@ucsb.edu)
"""

import os
import numpy as np

from skimage import exposure, morphology, segmentation, io
from scipy import ndimage, signal

import sep
from astropy.table import Table
from astropy.stats import sigma_clipped_stats

from cosmic_conn.cr_pipeline.zscale import zscale
from cosmic_conn.cr_pipeline.banzai_stats import (
    robust_standard_deviation,
    sigma_clipped_mean,
)


def mkdir(path):
    os.makedirs(path, exist_ok=True)


def subtract_sky(ndarray, remove_negative=False):
    if remove_negative:
        # subtracted = np.where(subtracted < 0, 0, subtracted)
        ndarray[ndarray < 0.0] = 0.0

    if len(ndarray.shape) == 3:
        median = np.median(ndarray, axis=(1, 2), keepdims=True)
        ndarray -= median
    elif len(ndarray.shape) == 2:
        ndarray -= np.median(ndarray)

    return ndarray.astype("float32")


def variation_intensity(vector):
    # modified from coefficient of variation
    # this function is better at spotting sudden bright pixel in a vector (CR)
    # https://en.wikipedia.org/wiki/Coefficient_of_variation

    # subtract negative numbers, minimize standard deviation
    if vector.min() < 0:
        vector -= vector.min()

    eps = 1
    median = np.median(vector)

    variation_intensity = (np.max(vector) - median) / (median + eps)

    return variation_intensity


def prune_nans_from_table(table):
    """
    From banzai
    https://github.com/LCOGT/banzai/blob/master/banzai/utils/array_utils.py
    """
    nan_in_row = np.zeros(len(table), dtype=bool)
    for col in table.colnames:
        nan_in_row |= np.isnan(table[col])
    return table[~nan_in_row]


def sep_source_mask(frames, valid_mask):
    """
    Following Banzai's SEP procedure with some modifications
    https://github.com/LCOGT/banzai/blob/master/banzai/photometry.py
    SEP manual
    https://www.astromatic.net/pubsvn/software/sextractor/trunk/doc/sextractor.pdf

    1. Extrat sources from the median frame, more reliable than SEP on individual frames
    2. Calculate each soruce's SEP stats, ellipse, window size
    3. Expand sources' rectangle window, acquire a local background mask
    4. Local background mask shuold be very close across frames, except for extreme PSF wings
    5. Calculate local background stats for each source and extract sources' mask from background
    6. Generate a source mask for each frame
    """

    # lower threshold as we work on the median frame
    threshold = 3.5
    min_area = 9

    # prep median frame for SEP source extractor
    median_frame = np.median(frames, axis=0)

    # increase limit as we work on the median frame
    ny, nx = median_frame.shape
    sep.set_extract_pixstack(int(nx * ny * 0.10))

    # subtract background
    bkg = sep.Background(median_frame, bw=32, bh=32, fw=3, fh=3)
    bkg_img = bkg.back()
    median_frame -= bkg_img

    # SEP applies a gaussian blur kernel
    # [[1,2,1], [2,4,2], [1,2,1]] by default
    try:
        # corner case when active sources pixel exceeds limit will crash
        sources = sep.extract(
            median_frame, threshold, err=bkg.globalrms, minarea=min_area
        )
    except:
        return None

    # Convert the detections into a table
    sources = Table(sources)

    if len(sources) > 3000:
        return None

    # We remove anything with a detection flag >= 8
    # This includes memory overflows and objects that are too close the edge
    sources = sources[sources["flag"] < 8]

    sources = prune_nans_from_table(sources)

    # Calculate the ellipticity and elongation
    sources["ellipticity"] = 1.0 - (sources["b"] / sources["a"])
    sources["elongation"] = sources["a"] / sources["b"]

    # Fix any value of theta that are invalid due to floating point rounding
    # -pi / 2 < theta < pi / 2
    sources["theta"][sources["theta"] > (np.pi / 2.0)] -= np.pi
    sources["theta"][sources["theta"] < (-np.pi / 2.0)] += np.pi

    # Calculate the FWHMs of the stars:
    fwhm = 2.0 * (np.log(2) * (sources["a"]
                               ** 2.0 + sources["b"] ** 2.0)) ** 0.5
    sources["fwhm"] = fwhm

    # small fwhm are often bad pixels/dust on CCD (consistent across frames)
    sources = sources[fwhm > 1.0]

    """
    Generate local mask for each source in each frame
    """
    sources_masks = []

    img_height = median_frame.shape[0]
    img_width = median_frame.shape[1]

    for frame in frames:
        sources_mask = np.zeros_like(median_frame, dtype="uint8")

        for i in range(len(sources)):
            src = sources[i]

            src_ellipse = np.zeros_like(median_frame, dtype="uint8")

            # soruce's ellipse based on the median frame
            sep.mask_ellipse(
                src_ellipse, src["x"], src["y"], src["a"], src["b"], src["theta"], r=2
            )

            # expand source window by 1.5x so more sky and PSF wings could be included
            x_expand = int((src["xmax"] - src["xmin"]) * 0.25)
            y_expand = int((src["ymax"] - src["ymin"]) * 0.25)

            # min, max to handle boundary cases
            ymin = max(0, src["ymin"] - y_expand)
            ymax = min(src["ymax"] + y_expand, img_height)
            xmin = max(0, src["xmin"] - x_expand)
            xmax = min(src["xmax"] + x_expand, img_width)

            # fail safe #0, if source too close to valid_mask boundary, pass
            # we got valid mask from reprojecting consecutive frames to same coordinates
            # so some frame will have missing information near boundary
            # valid_count = np.sum(valid_mask[ymin:ymax, xmin:xmax] == 0)
            # if valid_count > 0:
            #     continue

            src_window = np.zeros_like(median_frame, dtype="uint8")
            src_window[ymin:ymax, xmin:xmax] = 1

            # remove soruce ellipse from source window mask
            bkg_mask = src_window * (1 - src_ellipse)

            # fail safe #1, if ellipse larger than window, use ellipse directly
            if bkg_mask.sum() == 0.0:
                sources_mask[src_ellipse > 0] = 1
                continue

            # window crop background pixels from each frame
            bkg_frm = bkg_mask * frame

            # shift background and source pxiels values
            # valid bkg area will be above zero, all other area == 0
            zero_offset = np.min(bkg_frm) - 1
            src_frm = (frame - zero_offset) * src_window
            bkg_frm = (bkg_frm - zero_offset) * bkg_mask

            # if effective background usable, get local background stats
            # consider Lucy smoothing
            bkg_vec = bkg_frm[ymin:ymax, xmin:xmax].flatten()
            bkg_vec = bkg_vec[bkg_vec != 0]

            # use 2 sigma above local background median as the spread boundary
            # robust standard deviation ensures to reject PFS wings from stats
            # 2 sigma widen the source's extent slightly but includes more soft contour
            bkg_std = robust_standard_deviation(bkg_vec)
            bkg_upperbound = np.median(bkg_vec) + 2.0 * bkg_std

            src_peak_value = np.max(src_frm)
            # src_peak_value = max(src_peak, np.max(src_frm))

            # fail safe #2, if background upperbound higher than source peak
            if bkg_upperbound >= src_peak_value:
                sources_mask[src_ellipse > 0] = 1
                continue

            # effective pixel values range for the source
            tolerance = src_peak_value - bkg_upperbound

            # starting point for floodfill, from true peak pixel
            seed_point = np.unravel_index(src_frm.argmax(), src_frm.shape)
            # seed_point = (src['ypeak'], src['xpeak'])

            # flood filling is better than thresholding
            # as the mask expands from source center, CR in the source window won't be flagged
            src_mask = segmentation.flood(
                src_frm, seed_point, tolerance=tolerance
            ).astype("uint8")

            sources_mask[src_ellipse > 0] = 1
            sources_mask[src_mask > 0] = 1

        sources_masks.append(sources_mask)

    sources_masks = np.stack(sources_masks)

    # dilate CR by n pixel to avoid introducing artificl sharp edges
    struct = ndimage.generate_binary_structure(2, 2)
    sources_dilation = np.zeros_like(sources_masks)

    for i in range(sources_masks.shape[0]):
        sources_dilation[i] = ndimage.binary_dilation(
            sources_masks[i], structure=struct, iterations=2
        )

    return sources_dilation


def sigma_clip(ndarray, sigma=0.0, clip_source=False):
    # clamp extreme outlier pixels for better stats
    if sigma == 0.0:
        return ndarray

    else:
        mean, median, std = sigma_clipped_stats(ndarray, sigma=sigma)

        if clip_source:
            # clip extreme bright pixels
            ceil = mean + sigma * std
            ndarray[ndarray > ceil] = ceil

        # clip extreme low (negative) pixels
        floor = mean - sigma * std
        ndarray[ndarray < floor] = median

        return ndarray


def extract_CR(frames, median_frm, cr_mask, root, filenames):
    # extract actual CR values and output masks for data augmentation

    # frame is aleady sky subtracted
    cr_frames = frames * cr_mask

    # save all cr frames as npy
    out_dir = os.path.join(root, "cr_frames")
    mkdir(out_dir)
    for i in range(len(cr_frames)):
        fname = os.path.join(out_dir, filenames[i] + "_cr")
        np.savez_compressed(fname, cr_frames[i], allow_pickle=True)

    # export cleaned frames for png preview
    clean_frames = np.where(cr_mask > 0, median_frm, frames)

    return clean_frames, cr_frames


def replace_boundary(array, boundary, replace_value=0):
    array[0:boundary, :] = replace_value
    array[-boundary:, :] = replace_value
    array[:, 0:boundary] = replace_value
    array[:, -boundary:] = replace_value
    return array


def erase_boundary_np(ndarray, boundary_width=0, replace_value=0):
    # replace boundary with 0 to avoid occasional issues on CCD boundary
    if boundary_width == 0:
        return ndarray

    if boundary_width > 0:
        copy = ndarray.copy()

        if len(copy.shape) == 2:
            copy = replace_boundary(copy, boundary_width, replace_value)
            return copy

        elif len(copy.shape) == 3:
            for i in range(copy.shape[0]):
                copy[i] = replace_boundary(
                    copy[i], boundary_width, replace_value)
            return copy


def acquire_valid_area(valid_masks):
    # reprojection shifted/rotated frames in a sequence
    # leveing black pixels near edges, valid_masks are binary
    valid_mask = np.ones_like(valid_masks[0])

    for i in range(valid_masks.shape[0]):
        valid_mask *= valid_masks[i]

    assert valid_mask.max() <= 1.0, "valid mask max over 1."

    return valid_mask.astype("uint8")


def center_crop_npy(array, crop_size=0, crop_ratio=1):
    # center cropping a 2D array
    h, w = array.shape
    if crop_size == 0 and crop_ratio == 1:
        return array

    if crop_size > 0:
        if h < crop_size or w < crop_size:
            print(f"array smaller than crop size")
            return array
        else:
            left, top = int(w / 2 - crop_size / 2), int(h / 2 - crop_size / 2)
            return array[top: top + crop_size, left: left + crop_size]

    if 0 < crop_ratio < 1:
        new_w, new_h = int(w * crop_ratio), int(h * crop_ratio)
        w_start, h_start = int((w - new_w) / 2), int((h - new_h) / 2)
        return array[h_start: h_start + new_h, w_start: w_start + new_w]


def median_cut_off(img, n_std=3.0):
    img -= np.median(img)
    upper_bound = sigma_clipped_mean(img, sigma=3.5) + n_std * np.std(img)

    img = np.where(img < 0, 0, img)
    img = np.where(img > upper_bound, upper_bound, img)

    # linear normalize to increase contrast
    return img


def histogram_equalization(img, adaptive=False):
    img -= np.median(img) + 2 * robust_standard_deviation(img)
    img = np.where(img < 0, 0, img)

    # Contrast stretching
    p2, p98 = np.percentile(img, (2, 98))
    img_rescale = exposure.rescale_intensity(img, in_range=(p2, p98))

    if adaptive:
        # Adaptive Equalization
        img_eq = exposure.equalize_adapthist(img_rescale, clip_limit=0.03)
    else:
        # Equalization
        img_eq = exposure.equalize_hist(img_rescale)

    # scale to uint8 data range and data type
    img_uint8 = scale2uint8(img_eq)

    return img_uint8


def zscale_cutoff(array, z1z2=None, nsamples=600, contrast=0.25):
    # 600 samples, contrast 0.25, same as DS9

    if array.max() < 250:
        return array, z1z2

    if z1z2 is None:
        z1, z2 = zscale(array, nsamples, contrast)
        z1z2 = [z1, z2]
    else:
        z1, z2 = z1z2

    # array = np.where(array < z1, z1, array)
    # array = np.where(array > z2, z2, array)
    array[array < z1] = z1
    array[array > z2] = z2
    array -= array.min()

    return array, z1z2


def zscale_stack_cutoff(noisy_stack, nsamples=10_000, contrast=0.1, cut_offs=None):
    if cut_offs is None:
        # normalize all noisy frames in a stack to same level
        cut_offs = [zscale(nsy, nsamples, contrast) for nsy in noisy_stack]

    z1 = min([c[0] for c in cut_offs])
    z2 = max([c[1] for c in cut_offs])

    noisy_vis = np.clip(noisy_stack, z1, z2)

    return noisy_vis, cut_offs


def scale2uint8(array):
    array_new = array - array.min()
    assert (
        array_new.max() != 0
    ), f"zero division when normalizing image, array max == 0 {array.shape}"

    array_new /= array_new.max()
    array_new = (array_new * 255).astype("uint8")
    return array_new


def save_as_png(
    array,
    dir,
    filename,
    file_type="uint8",
    zscale=False,
    z1z2=None,
    gamma=False,
    equalization=False,
    grid=False,
):

    if not isinstance(array, np.ndarray):
        array = np.array(array)

    # array = array.astype("int") if "scrappy" in filename else array
    array = array * 255.0 if "mask" in file_type else array

    hdr = False
    for key in ["noisy", "clean", "cleaned", "median"]:
        if key in file_type:
            hdr = True

    if hdr:
        if zscale:
            # 600 samples, contrast 0.25, same as DS9
            array, z1z2 = zscale_cutoff(
                array, z1z2, nsamples=600, contrast=0.25)

        if gamma:
            array -= array.min()
            # STIFF default gamma 2.2
            array = exposure.adjust_gamma(array, 2.2)

        if equalization:
            file_type += "_equal_"
            array = histogram_equalization(array, adaptive=False)

        if not (zscale or gamma or equalization):
            array = median_cut_off(array, n_std=2)

    # scale to uint8 data range and data type
    array = scale2uint8(array)

    if grid:
        draw_grid(array, interval=100, grid_color=32)

    # save as uint8 image
    mkdir(dir)
    filename = filename + "_" + file_type + ".png"
    path = os.path.join(dir, filename)
    io.imsave(path, array, check_contrast=False)

    return filename, z1z2


def draw_grid(array, interval=100, grid_color=1):

    # Modify the image to include the grid
    array[:, interval::interval] = grid_color
    array[interval::interval, :] = grid_color

    return array



# -*- coding: utf-8 -*-

"""
Boning Dong, CY Xu (cxu@ucsb.edu)
"""

import math
import numpy as np


def find_noise_thumbnail_coords(mask, threshold=0.5, patch_size=32):
    """
    Return CR thumbnails's top-left corner coordinate
    """
    confidence_coord_tuples = []  # the coords have format [x, y]
    border = 64

    shape = mask.shape
    hh = math.ceil((shape[0] - 2 * border) // patch_size)
    ww = math.ceil((shape[1] - 2 * border) // patch_size)

    for i in range(hh):
        for j in range(ww):
            # overlapping crop at stride
            h_start = border + i * patch_size
            w_start = border + j * patch_size

            if h_start + patch_size > (shape[0] - border):
                h_start = shape[0] - border - patch_size

            if w_start + patch_size > (shape[1] - border):
                w_start = shape[1] - border - patch_size

            h_stop = h_start + patch_size
            w_stop = w_start + patch_size

            patch = mask[h_start:h_stop, w_start:w_stop]
            # np.where(patch > threshold, 1, 0)

            confidence_coord_tuples.append((patch.sum(), [w_start, h_start]))

    confidence_coord_tuples.sort(key=lambda elem: elem[0], reverse=True)

    return [x[1] for x in confidence_coord_tuples]



# -*- coding: utf-8 -*-

"""
Tools to support consoles commands and app.
CY Xu (cxu@ucsb.edu)
"""

import os
import argparse
import glob
import requests
import zipfile
from pathlib import Path
from cosmic_conn.dl_framework.options import ModelOptions

PYTEST_DATA_URL = "https://sites.cs.ucsb.edu/~cy.xu/cosmic_conn/Cosmic-CoNN_test_data.zip"
MODEL_URL = "https://sites.cs.ucsb.edu/~cy.xu/cosmic_conn/trained_models.zip"

EXTENSIONS = ["fits", "fz"]

def is_fits_file(filename):
    filename = filename.lower()
    return any(filename.endswith(extension) for extension in EXTENSIONS)

def download_file(url):
    local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                f.write(chunk)
    return local_filename


def download_test_data():
    """ download three banzai reduced FITS for tests
    """
    test_data_path = os.path.join(os.getcwd(), "Cosmic-CoNN_test_data")
    print(f"Downloading Cosmic-CoNN test data to {test_data_path}")
    file_name = download_file(PYTEST_DATA_URL)

    with zipfile.ZipFile(file_name, "r") as zip_ref:
        zip_ref.extractall(test_data_path)

    assert os.path.exists(test_data_path)
    os.remove(file_name)

    return test_data_path


def download_trained_models(model_path):
    print(f"Downloading trained Cosmic-CoNN models to {model_path}")
    model_name = download_file(MODEL_URL)

    with zipfile.ZipFile(model_name, "r") as zip_ref:
        zip_ref.extractall(model_path)

    assert os.path.exists(model_path)
    os.remove(model_name)

    return model_path


def check_trained_models():
    parent_dir = str(Path(__file__).parent.absolute())
    model_dir = os.path.join(parent_dir, "trained_models")
    models = glob.glob(model_dir + "/*.pth")
    if len(models) < 3:
        raise ValueError("models not found in trained_models directory.")
    return model_dir


def console_arguments():
    """Parse console arguments for cosmic_conn."""
    parser = argparse.ArgumentParser(
        description="Cosmic-CoNN console commands to detect CRs "
        "in FITS file(s). CR prediction will be attached "
        "as a FITS extention. A new copy of the FITS "
        "file is saved in cosmic_conn_output."
    )
    parser.add_argument(
        "-a", "--app", action="store_true", help="launch the CR "
        "detector web app"
    )
    parser.add_argument(
        "-i",
        "--input",
        default="./",
        help="input FITS file or directory for CR detection. "
        "Passing a directory will batch detect all FITS files,"
        "including subdirectories.",
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        default="ground_imaging",
        help="use one of the models: ground_imaging | NRES | HST_ACS_WFC",
    )
    parser.add_argument(
        "-e",
        "--ext",
        type=str,
        default="SCI",
        help="read data from this hdul extension, SCI by default."
    )

    opt = parser.parse_args()
    opt = vars(opt)

    # swtich to custom class from argparse dict
    model_options = ModelOptions()
    for k, v in opt.items():
        model_options[k] = v

    return model_options


def parse_input(input, PREDICT_DIR):
    all_fits = []

    # sanity check
    assert os.path.exists(input), (
        f"{input} does not exist. Usage: cosmic-conn --help"
    )

    # a single file or files in directory?
    if os.path.isfile(input):
        all_fits.append(input)

    elif os.path.isdir(input):
        # walk through all subdirectories to find all FITS
        for root, dirs, files in os.walk(input):
            if PREDICT_DIR in root:
                continue

            files = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]

            # check extention for FITS file
            files = [f for f in files if is_fits_file(f)]
            files = [os.path.join(root, f) for f in files]

            all_fits += files

        print(f"{len(all_fits)} FITS found in {input}")

    assert len(all_fits) > 0, (
        (f"Invalid input {input}. Use -i INPUT to specify "
            "FITS file or a directory path")
    )

    return all_fits


# if __name__ == "__main__":
#     download_test_data()
#     check_trained_models()



# -*- coding: utf-8 -*-

"""
Main file to define the CR labeling pipelien for LCO NRES data
CY Xu (cxu@ucsb.edu)
"""

import os
import numpy as np
from astropy.io import fits

from skimage import morphology
from scipy import ndimage

from cosmic_conn.cr_pipeline.utils_io import save_as_png, save_array_png
from cosmic_conn.cr_pipeline.utils_io import remove_outlier_frm, remove_ext, mkdir


class NRES_CR_detector:
    def __init__(self, hdul=None, opt=None):
        self.hdul = hdul
        self.opt = opt

        header = self.hdul["SPECTRUM"].header

        self.fname = header["filename"]
        self.exp_time = header["exptime"]
        self.rdnoise = header["rdnoise"]
        self.gain = header["gain"]
        maxlin = header["maxlin"]

        self.min_cr_size = self.opt.min_cr_size

        frames, self.filenames = [], []
        for i in range(1, 4):
            frames.append(hdul[i].data)
            self.filenames.append(hdul[i].header["filename"])

        # 0. maxlin header cutoff
        frames = np.stack(frames).astype("float32")
        frames[frames > maxlin] = maxlin

        # 3. saturated pixel radius 3 dilation to ignore
        ignore_mask = (frames >= maxlin).astype("uint8")
        struct = ndimage.generate_binary_structure(2, 2)

        for i in range(ignore_mask.shape[0]):
            ignore_mask[i] = ndimage.binary_dilation(
                ignore_mask[i], structure=struct, iterations=3
            )
        self.ignore_mask = ignore_mask

        # 1. subtract overscan (right hand)
        overscan = frames[:, :, -35:]
        frames -= np.median(overscan, axis=(1, 2), keepdims=True)

        # 2. multiply gain
        self.frames = frames * self.gain
        # self.rdnoise *= self.gain

        # pre-processing, a separate copy for mask generation
        # self.frames = sigma_clip(self.frames, sigma=3.0, clip_bright_source=False)

        self.reject_dir = self.opt.root + "/rejected_frms"

    def generate_CR_masks(self):
        """
        This is the main CR detection loop
            estimate_reject_refine() is called one or multiple times for CR detection
            detect CR in the sequence of frames, reject outlier pixels and frames

        if estimate_reject_refine() returns a string, it's a error message during detection
        print the error message and abort this sequence
        """
        high_thres = self.opt.snr_thres
        low_thres = self.opt.snr_thres_low

        # initial estimation to reject outlier frame
        cr_frac = self.estimate_reject_refine(high_thres, low_thres)

        if isinstance(cr_frac, str):
            self.output_png_preview(reason=cr_frac)
            return cr_frac

        # if an outlier frame is rejected, detect again with the new sequence because the
        # CR frac between frames might change, until all frames considered well aligned
        # while len(self.reject_idx) > 0:
        #     cr_frac = self.estimate_reject_refine(high_thres, low_thres)
        #     if isinstance(cr_frac, str): return cr_frac

        # all rejection passed, print stats of the useable frames
        print(
            f"{self.fname}, exp_time {self.exp_time}, {[np.sum(m) for m in self.cr_mask]}\n",
            f"SNR {high_thres}, {low_thres}, CR fraction {round(cr_frac, 6)}\n",
        )

        # append CR mask
        for i in range(3):
            self.hdul.append(
                fits.CompImageHDU(
                    self.cr_mask[i],
                    header=self.hdul[i].header,
                    name="CR",
                    ver=i,
                    uint=True,
                )
            )

        for i in range(3):
            self.hdul.append(
                fits.CompImageHDU(
                    self.ignore_mask[i],
                    header=self.hdul[i].header,
                    name="IGNORE",
                    ver=i,
                    uint=True,
                )
            )

        new_path = os.path.join(self.opt.data, "masked_fits")
        new_filename = f"{remove_ext(self.fname)}_masks.fits"

        # save aligned fits to sub-directory
        mkdir(new_path)
        out_path = os.path.join(new_path, new_filename)

        self.hdul.writeto(out_path, overwrite=True)

        # output png preview or numpy as training set (no longer using numpy)
        if not self.opt.no_png:
            self.clean_frames = np.where(
                self.cr_mask > 0, self.median_frame, self.frames
            )
            self.output_png_preview(self.opt.comment)

    # @profile
    def estimate_reject_refine(self, high_thres, low_thres, reject_frm=True):
        """
        The main CR detection happens here:
            1. estimate two masks based on an initial SNR threshold and
               a lower threshold (for peripheral CR edges)

            4. refine CR to incldue soft edges via dilation/flood filling

            7. reject frame outliers with excessive CR in a sequence
               often false positives caused by satellite trails or mis-aligned frames 
        """

        # 1.
        mask_low, mask_high = self.estimate_two_masks(
            self.frames, self.rdnoise, high_thres, low_thres
        )
        # from here on, self.frames are sky subtracted

        # 4.
        self.cr_mask = self.maximize_cr_edges(
            mask_low, mask_high, self.min_cr_size)

        # return error message
        if isinstance(self.cr_mask, str):
            return self.cr_mask

        # 7.
        # if reject_frm:
        #     self.reject_idx = self.reject_outlier_frm()
        # else:
        #     self.reject_idx = []

        # calculate CR fraction/ratio, used for adaptive SNR threshold
        total_pixels = self.cr_mask.size
        cr_frac = np.sum(self.cr_mask) / total_pixels

        return cr_frac

    def compute_SNR(self, frames, read_noise):
        # assuming frames in an N_images x Ny x Nx array
        N_frames = frames.shape[0]

        bkg_noise = np.median(frames, axis=(1, 2), keepdims=True)

        # calculate uncertainties
        frame_uncertainty = np.sqrt(
            read_noise ** 2 + np.abs(frames) + bkg_noise)
        # frame_uncertainty = np.sqrt(read_noise**2 + np.abs(frames))

        median_frame_uncertainty = np.median(frame_uncertainty, axis=0) / np.sqrt(
            N_frames
        )

        # calculate the total noise including our model (approximately for the median)
        total_uncertainty = np.sqrt(
            frame_uncertainty ** 2 + median_frame_uncertainty ** 2
        )

        # subtract sky and get correct median
        # frames = subtract_sky(frames, remove_negative=False)
        self.median_frame = np.median(frames, axis=0)

        # The median acts as our model of what the frame should look like,
        # so we can look for deviations from that
        # without np.abs(frames), only consider positive deviations
        snr_deviation = (frames - self.median_frame) / total_uncertainty

        return snr_deviation

    def estimate_two_masks(self, frames, read_noise, high_thres, low_thres):
        """
        noisy_stack: ndarray
            a stack of noisy frames of shape (n, h, w)

        Rerturn:
            mask_low, retrieved from a lower threshold, more pixel detected but more false positive as well
            mask_high, retrieved from a higher threshold to reject false positive
        """

        snr_deviation = self.compute_SNR(frames, read_noise)

        mask_high = (snr_deviation > high_thres).astype("uint8")

        # lower threshold to get the soft edges of CR
        mask_low = (snr_deviation > low_thres).astype("uint8")

        # for debugging
        self.snr_deviation = snr_deviation
        self.mask_high, self.mask_low = mask_high, mask_low

        return mask_low, mask_high

    def maximize_cr_edges(self, mask_low, mask_high, min_cr_size):
        shape = mask_high.shape
        cr_mask = np.zeros_like(mask_low, dtype="uint8")

        # dilation expands CR edges by N pixels, more conservative than flood filling
        # current 3 iterations will expand high mask by 3 pixels

        mask_high *= 1 - self.ignore_mask

        struct = ndimage.generate_binary_structure(2, 2)
        mask_dilation = np.zeros_like(mask_high)

        for i in range(shape[0]):
            mask_dilation[i] = ndimage.binary_dilation(
                mask_high[i], structure=struct, iterations=self.opt.dilation
            )

        cr_mask = np.logical_and(mask_dilation, mask_low).astype("uint8")

        cr_mask *= 1 - self.ignore_mask

        # save before removing single pixels to add them to ignore mask
        cr_mask_single_pixels = cr_mask.copy().astype("uint8")

        # remove flagged CR smaller than certain pixels from final mask
        if min_cr_size > 1:
            for i in range(shape[0]):
                # create labels in segmented image
                imglab = morphology.label(cr_mask[i], connectivity=2)
                # remove small patches less than min_cr_size
                cr_mask[i] = morphology.remove_small_objects(
                    imglab, min_size=min_cr_size, connectivity=2
                )

        cr_mask = (cr_mask > 0).astype("uint8")

        # add single pixels and source mask to make the ignore mask
        self.single_pixels = cr_mask_single_pixels - cr_mask
        self.ignore_mask = (
            (self.ignore_mask + self.single_pixels) > 0).astype("uint8")

        cr_mask *= 1 - self.ignore_mask

        return cr_mask

    def reject_outlier_frm(self):
        shape = self.cr_mask.shape

        # cr counts in each frame
        cr_counts = np.sum(self.cr_mask, axis=(1, 2))

        total_cr = 1 if np.sum(cr_counts) == 0 else np.sum(cr_counts)
        frame_cr_contribution = cr_counts / total_cr

        # if the # of CR from one frames is more than any of the two conditions
        #   1. twice the average,
        #   2. two standard deviation of avergae contribution (outlier)
        # reject this frame, as it might have satellite trail, mis-aligned, or blurry

        mean = np.mean(frame_cr_contribution)
        reject_thres = min(2.0 * mean, mean + 2.0 *
                           np.std(frame_cr_contribution))

        reject_idx = np.where(frame_cr_contribution > reject_thres)[0].tolist()

        # save rejected frames for reference
        if len(reject_idx) > 0:

            for i in range(shape[0]):
                name = "noisy_rejected" if i in reject_idx else "noisy_accepted"
                save_as_png(
                    self.frames[i],
                    self.reject_dir,
                    self.filenames[i],
                    name,
                    zscale=True,
                )

            print(f"rejected {self.fname}, reject_thres {reject_thres}")
            print(frame_cr_contribution)
            print()

        # remove rejected frames from frame stack and fits
        if len(reject_idx) > 0:
            self.frames = remove_outlier_frm(self.frames, reject_idx)
            self.frmtotal = self.frames.shape[0]

            self.frames_banzai = remove_outlier_frm(
                self.frames_banzai, reject_idx)

            self.cr_mask = remove_outlier_frm(self.cr_mask, reject_idx)

            self.filenames = remove_outlier_frm(
                np.array(self.filenames), reject_idx
            ).tolist()

            self.hdul = remove_outlier_frm(
                self.hdul, reject_idx, update_hdul=True)

        return reject_idx

    def output_png_preview(self, reason):
        if "reject" in reason:
            save_array_png(
                self.frames_banzai,
                None,
                None,
                None,
                self.opt.root,
                self.filenames,
                save_np=False,
                save_png=True,
                header=self.hdul[1].header,
                comment=reason,
            )
        else:
            save_array_png(
                self.frames,
                self.cr_mask,
                self.clean_frames,
                None,
                self.opt.root,
                self.filenames,
                save_np=False,
                save_png=True,
                header=self.hdul[1].header,
                comment=reason,
            )
